# -*- coding: utf-8 -*
"""Convert ISA-Tab to ISA-JSON"""
import glob
import json
import logging
import os
import re
from enum import Enum
from os.path import join
from uuid import uuid4

from jsonschema import Draft4Validator, RefResolver

from isatools import isatab
from isatools.io.isatab_parser import parse
from isatools.isajson import ISAJSONEncoder


log = logging.getLogger('isatools')

SCHEMAS_PATH = join(os.path.dirname(os.path.realpath(__file__)),
                    "../resources/schemas/isa_model_version_1_0_schemas/core/")
INVESTIGATION_SCHEMA = "investigation_schema.json"

# REGEXES
_RX_COMMENTS = re.compile(r'Comment\[(.*?)\]')


class IdentifierType(Enum):
    counter = 1
    uuid = 2
    name = 3


def convert(work_dir, identifier_type=IdentifierType.name,
            validate_first=True, config_dir=isatab.default_config_dir,
            use_new_parser=False):
    i_files = glob.glob(os.path.join(work_dir, 'i_*.txt'))
    if validate_first:
        log.info("Validating input ISA tab before conversion")
        if len(i_files) != 1:
            log.fatal("Could not resolve input investigation file, "
                      "please check input ISA tab directory")
            return
        with open(i_files[0], 'r', encoding='utf-8') as validate_fp:
            report = isatab.validate(fp=validate_fp, config_dir=config_dir,
                                     log_level=logging.ERROR)
            if len(report['errors']) > 0:
                log.fatal("Could not proceed with conversion as there are some"
                          " fatal validation errors. Check log")
                return
    if use_new_parser:
        log.info("Using new ISA-Tab parser")
        log.info("Loading ISA-Tab: %s", i_files[0])
        with open(i_files[0], 'r', encoding='utf-8') as fp:
            ISA = isatab.load(fp)
            log.info("Dumping ISA-JSON")
            return json.loads(json.dumps(ISA, cls=ISAJSONEncoder))
    else:
        converter = ISATab2ISAjson_v1(identifier_type)
        log.info("Using old parser")
        return converter.convert(work_dir)


class ISATab2ISAjson_v1:

    MATERIAL_TYPE = "Material Type"
    LABEL = "Label"
    CHARACTERISTICS = "Characteristics"
    CHARACTERISTIC_CATEGORY = "characteristic_category"
    FACTOR_VALUE = "Factor Value"
    UNIT = "Unit"
    PARAMETER_VALUE = "Parameter Value"
    ARRAY_DESIGN_REF = "Array Design REF"

    def __init__(self, identifier_type):
        self.identifiers = list()  # list of dictionaries
        self.counters = dict()
        self.identifier_type = identifier_type

    def setIdentifier(self, type, name, identifier):
        self.identifiers.append(dict([("type", type), ("name", name),
                                      ("identifier", identifier)]))

    def getIdentifier(self, type, name):
        for subVal in self.identifiers:
            if subVal["type"] == type and subVal["name"] == name:
                return subVal["identifier"]

    def generateIdentifier(self, type, name):
        try:
            self.counters[type] += 1
        except KeyError:
            self.counters[type] = 1

        identifier = ""

        if self.identifier_type == IdentifierType.counter:
            identifier = \
                "http://data.isa-tools.org/"+type+"/"+str(self.counters[type])
        elif self.identifier_type == IdentifierType.uuid:
            identifier = "http://data.isa-tools.org/UUID/"+str(uuid4())
        elif self.identifier_type == IdentifierType.name:
            identifier = "#"+type+"/"+name.replace(" ", "_")

        self.setIdentifier(type, name, identifier)
        return identifier

    def convert(self, work_dir):
        """
            Convert an ISA-Tab dataset (version 1) to JSON provided the ISA
            model v1.0 JSON Schemas
            :param work_dir: directory containing the ISA-tab dataset
        """
        log.info("Converting ISA-Tab to ISA-JSON for %s", work_dir)

        isa_tab = parse(work_dir)

        if isa_tab is None:
            log.fatal("No ISA-Tab dataset found")
        else:
                isa_json = dict([])
                if isa_tab.metadata != {}:
                    isa_json = dict([
                        ("identifier",
                         isa_tab.metadata['Investigation Identifier']),
                        ("title", isa_tab.metadata['Investigation Title']),
                        ("description",
                         isa_tab.metadata['Investigation Description']),
                        ("submissionDate",
                         isa_tab.metadata['Investigation Submission Date']),
                        ("publicReleaseDate",
                         isa_tab.metadata[
                             'Investigation Public Release Date']),
                        ("ontologySourceReferences",
                         self.createOntologySourceReferences(
                             isa_tab.ontology_refs)),
                        ("publications",
                         self.createPublications(isa_tab.publications,
                                                 "Investigation")),
                        ("people", self.createContacts(isa_tab.contacts,
                                                       "Investigation")),
                        ("studies", self.createStudies(isa_tab.studies)),
                        ("comments", self.createComments(isa_tab.metadata))
                    ])

                # validate json
                with open(join(SCHEMAS_PATH, INVESTIGATION_SCHEMA)) as json_fp:
                    schema = json.load(json_fp)
                    resolver = RefResolver(
                        'file://' + join(SCHEMAS_PATH, INVESTIGATION_SCHEMA),
                        schema)
                    validator = Draft4Validator(schema, resolver=resolver)
                    validator.validate(isa_json, schema)

                    log.info("Conversion finished")
                    return isa_json

    def createComments(self, isadict):
        comments = []
        for k in [k for k in isadict.keys() if _RX_COMMENTS.match(k)]:
            comments.append(
                self.createComment(_RX_COMMENTS.findall(k)[0], isadict[k]))
        return comments

    def createComment(self, name, value):
        comment_json = dict([
            ("name", name),
            ("value", value)
        ])
        return comment_json

    def createContacts(self, contacts, inv_or_study):
        people_json = []
        for contact in contacts:
            person_last_name = contact[inv_or_study+" Person Last Name"]
            if not person_last_name:
                continue
            person_identifier = self.generateIdentifier("person",
                                                        person_last_name)
            person_json = dict([
                ("@id", person_identifier),
                ("lastName", person_last_name),
                ("firstName", contact[inv_or_study+" Person First Name"]),
                ("midInitials", contact[inv_or_study+" Person Mid Initials"]),
                ("email", contact[inv_or_study+" Person Email"]),
                ("phone", contact[inv_or_study+" Person Phone"]),
                ("fax", contact[inv_or_study+" Person Fax"]),
                ("address", contact[inv_or_study+" Person Address"]),
                ("affiliation", contact[inv_or_study+" Person Affiliation"]),
                ("roles", self.createOntologyAnnotationsFromStringList(
                    contact, inv_or_study, " Person Roles")),
                ("comments", self.createComments(contact))
            ])

            people_json.append(person_json)
        return people_json

    def createPublications(self, publications, inv_or_study):
        publications_json = []
        for pub in publications:
            publication_json = dict([
                ("pubMedID", pub[inv_or_study+' PubMed ID']),
                ("doi", pub[inv_or_study+' Publication DOI']),
                ("authorList", pub[inv_or_study+' Publication Author List']),
                ("title", pub[inv_or_study+' Publication Title']),
                ("status", self.createOntologyAnnotationForInvOrStudy(
                    pub, inv_or_study, " Publication Status"))
            ]
            )
            publications_json.append(publication_json)
        return publications_json

    def createProtocols(self, protocols, assays):
        protocols_json = []

        protocols_to_attach_parameter = []
        #  keep protocols that should have ArrayDesignREF as a parameter
        for assay in assays:
            for process_node in assay.process_nodes.values():
                if self.ARRAY_DESIGN_REF in process_node.parameters:
                        protocols_to_attach_parameter.append(
                            process_node.protocol)
        protocol_identifier = self.generateIdentifier("protocol", "unknown")
        protocol_json = dict([
                                 ("@id", protocol_identifier),
                                 ("name", "unknown"),
                                 ("protocolType", dict([
                                     ("annotationValue", "")
                                 ])),
                                 ("description", ""),
                                 ("uri", ""),
                                 ("version", ""),
                                 ("parameters", []),
                                 ("components", [])
                             ])
        protocols_json.append(protocol_json)
        for protocol in protocols:
            protocol_name = protocol['Study Protocol Name']
            if not protocol_name:
                continue
            protocol_identifier = self.generateIdentifier(
                "protocol", protocol_name)
            parameters = self.createProtocolParameterList(protocol)

            protocol_json = dict([
                ("@id", protocol_identifier),
                ("name", protocol_name),
                ("protocolType", self.createOntologyAnnotationForInvOrStudy(
                    protocol, "Study", " Protocol Type")),
                ("description", protocol['Study Protocol Description']),
                ("uri", protocol['Study Protocol URI']),
                ("version", protocol['Study Protocol Version']),
                ("parameters", parameters),
                ("components", self.createProtocolComponentList(protocol))
                ])
            protocols_json.append(protocol_json)

        return protocols_json

    def createProtocolParameterList(self, protocol):
        json_list = []
        parameters_json = self.createOntologyAnnotationsFromStringList(
            protocol, "Study", " Protocol Parameters Name")
        i = 0
        for parameter_json in parameters_json:
            parameter_identifier = self.generateIdentifier(
                "parameter", parameters_json[i]["annotationValue"])
            json_item = dict([
                ("@id", parameter_identifier),
                ("parameterName",  parameter_json)
            ])
            json_list.append(json_item)
            i += 1
        return json_list

    def createOntologyAnnotationForInvOrStudy(
            self, object, inv_or_study, type):
        onto_ann = dict([
                ("annotationValue", object[inv_or_study+type]),
                ("termSource", object[inv_or_study+type+" Term Source REF"]),
                ("termAccession", object[
                    inv_or_study+type+" Term Accession Number"])
        ])
        return onto_ann

    def createOntologyAnnotation(self, name, termSource, termAccesssion):
        onto_ann = dict([
            ("annotationValue", name),
            ("termSource", termSource),
            ("termAccession", termAccesssion)
        ])
        return onto_ann

    def createOntologyAnnotationsFromStringList(
            self, object, inv_or_study, type):
        name_array = object[inv_or_study+type].split(";")
        term_source_array = object[
            inv_or_study+type+" Term Source REF"].split(";")
        term_accession_array = object[
            inv_or_study+type+" Term Accession Number"].split(";")
        onto_annotations = []
        for i in range(0, len(name_array)):
            if (not name_array[i]):
                continue
            onto_ann = self.createOntologyAnnotation(name_array[i],
                                                     term_source_array[i],
                                                     term_accession_array[i])
            onto_annotations.append(onto_ann)
        return onto_annotations

    def createOntologyAnnotationListForInvOrStudy(
            self, array, inv_or_study, type):
        onto_annotations = []
        for object in array:
            onto_ann = self.createOntologyAnnotation(
                object[inv_or_study+type],
                object[inv_or_study+type+" Term Source REF"],
                object[inv_or_study+type+" Term Accession Number"])
            onto_annotations.append(onto_ann)
        return onto_annotations

    def createOntologySourceReferences(self, ontology_refs):
        ontologies = []
        for ontology_ref in ontology_refs:
            ontology = dict([
                ("description", ontology_ref["Term Source Description"]),
                ("file", ontology_ref["Term Source File"]),
                ("name", ontology_ref["Term Source Name"]),
                ("version", ontology_ref["Term Source Version"])
            ])
            ontologies.append(ontology)
        return ontologies

    def createStudies(self, studies):
        study_array = []
        for study in studies:
            study_name = study.metadata['Study Identifier']
            study_identifier = self.generateIdentifier("study", study_name)
            characteristics_categories_list = \
                self.createCharacteristicsCategories(study.nodes)
            unit_categories_list = self.createUnitsCategories(study.nodes)
            factors_list = self.createStudyFactorsList(study.factors)
            source_dict = self.createSourcesDictionary(study.nodes)
            sample_dict = self.createSampleDictionary(study.nodes)
            material_dict = self.createMaterialDictionary(study.nodes)
            protocol_list = self.createProtocols(study.protocols, study.assays)
            assay_list = self.createStudyAssaysList(study.assays, sample_dict)
            data_dict = self.createDataFiles(study.nodes)
            studyJson = dict([
                ("@id", study_identifier),
                ("identifier", study_name),
                ("title", study.metadata['Study Title']),
                ("description", study.metadata['Study Description']),
                ("submissionDate", study.metadata['Study Submission Date']),
                ("publicReleaseDate", study.metadata[
                    'Study Public Release Date']),
                ("studyDesignDescriptors",
                 self.createOntologyAnnotationListForInvOrStudy(
                     study.design_descriptors, "Study", " Design Type")),
                ("publications", self.createPublications(
                    study.publications, "Study")),
                ("people", self.createContacts(study.contacts, "Study")),
                ("protocols", protocol_list),
                ("factors", factors_list),
                ("characteristicCategories", characteristics_categories_list),
                ("unitCategories", unit_categories_list),
                ("materials", dict([
                        ("sources", list(source_dict.values())),
                        ("samples", list(sample_dict.values())),
                        ("otherMaterials", list(material_dict.values()))
                ])),
                ("processSequence", self.createProcessSequence(
                    study.process_nodes, source_dict, sample_dict,
                    material_dict, data_dict)),
                ("assays", assay_list),
                ("filename", study.metadata['Study File Name']),
                ("comments", self.createComments(study.metadata)),
            ])
            #  clean up unknown process if it's not used in process sequence
            unknown_used = False
            for assay in assay_list:
                for process in assay['processSequence']:
                    if process[
                            'executesProtocol']['@id'] == '#protocol/unknown':
                        unknown_used = True
                        break
            if not unknown_used:
                try:
                    unknown_prot_index = protocol_list.index(dict([
                                     ("@id", "#protocol/unknown"),
                                     ("name", "unknown"),
                                     ("protocolType", dict([
                                         ("annotationValue", "")
                                     ])),
                                     ("description", ""),
                                     ("uri", ""),
                                     ("version", ""),
                                     ("parameters", []),
                                     ("components", [])
                                 ]))
                    del studyJson['protocols'][unknown_prot_index]
                except ValueError:
                    pass
            study_array.append(studyJson)
        return study_array

    def createProtocolComponentList(self, protocol):
        json_list = []
        components_name = protocol['Study Protocol Components Name'].split(";")
        components_type_json = self.createOntologyAnnotationsFromStringList(
            protocol, "Study", " Protocol Components Type")
        index = 0
        for component_type_json in components_type_json:
            component_name = components_name[index]
            json_item = dict([
                ("componentName", component_name),
                ("componentType",  component_type_json)
            ])
            json_list.append(json_item)
            index += 1
        return json_list

    def createStudyFactorsList(self, factors):
        json_list = []
        for factor in factors:
            factor_identifier = self.generateIdentifier(
                "factor", factor['Study Factor Name'])
            json_item = dict([
                ("@id", factor_identifier),
                ("factorName", factor['Study Factor Name']),
                ("factorType", self.createOntologyAnnotation(
                    factor['Study Factor Type'],
                    factor['Study Factor Type Term Source REF'],
                    factor['Study Factor Type Term Accession Number']))
            ])
            json_list.append(json_item)
        return json_list

    def createProcessSequence(
            self, process_nodes, source_dict, sample_dict, material_dict,
            data_dict):
        json_list = []
        #  generate all the identifiers
        for process_node_name in process_nodes:
            self.generateIdentifier("process", process_node_name)

        for process_node_name in process_nodes:
            process_node = process_nodes[process_node_name]

            process_identifier = self.getIdentifier(
                "process", process_node_name)
            protocol_executed = self.createExecuteStudyProtocol(
                process_node_name, process_node)
            previous_process_identifier = \
                self.getIdentifier(
                    "process", process_node.previous_process.name) if \
                process_node.previous_process else ""
            next_process_identifier = \
                self.getIdentifier(
                    "process", process_node.next_process.name) if \
                process_node.next_process else ""
            if (process_node.assay_name):
                json_item = dict([
                    ("@id", process_identifier),
                    ("name", process_node.assay_name),
                    ("executesProtocol", protocol_executed),
                    ("performer", process_node.performer),
                    ("date", process_node.date),
                    ("parameterValues", self.createValueList(
                        self.PARAMETER_VALUE, process_node_name,
                        process_node)),
                    ("inputs", self.createInputList
                        (process_node.inputs, source_dict, sample_dict,
                            material_dict, data_dict)),
                    ("outputs", self.createOutputList(
                        process_node.outputs, sample_dict, material_dict,
                        data_dict)),
                    ("comments", self.createFromNodeComments(process_node)),
                ])
            else:
                json_item = dict([
                    ("@id", process_identifier),
                    ("executesProtocol", protocol_executed),
                    ("performer", process_node.performer),
                    ("date", process_node.date),
                    ("parameterValues", self.createValueList(
                        self.PARAMETER_VALUE, process_node_name,
                        process_node)),
                    ("inputs", self.createInputList(
                        process_node.inputs, source_dict, sample_dict,
                        material_dict, data_dict)),
                    ("outputs", self.createOutputList(
                        process_node.outputs, sample_dict, material_dict,
                        data_dict)),
                    ("comments", self.createFromNodeComments(process_node)),
                ])

            if previous_process_identifier:
                json_item.update({
                    "previousProcess": dict([(
                        "@id", previous_process_identifier)])
                })
            if next_process_identifier:
                json_item.update({
                    "nextProcess": dict([("@id", next_process_identifier)])
                })
            json_list.append(json_item)
        return json_list

    def createInputList(
            self, inputs, source_dict, sample_dict, material_dict, data_dict):
        json_list = []
        for argument in inputs:
            try:
                json_item = source_dict[argument]
                source_id = dict([("@id", json_item["@id"])])
                json_list.append(source_id)
            except KeyError:
                pass
            try:
                json_item = sample_dict[argument]
                sample_id = dict([("@id", json_item["@id"])])
                json_list.append(sample_id)
            except KeyError:
                pass
            try:
                json_item = material_dict[argument]
                material_id = dict([("@id", json_item["@id"])])
                json_list.append(material_id)
            except KeyError:
                pass

            try:
                json_item = data_dict[argument]
                data_id = dict([("@id", json_item["@id"])])
                json_list.append(data_id)
            except KeyError:
                pass
        return json_list

    def createOutputList(
            self, arguments, sample_dict, material_dict, data_dict):
        json_list = []
        for argument in arguments:
            try:
                json_item = sample_dict[argument]
                sample_id = dict([("@id", json_item["@id"])])
                json_list.append(sample_id)
            except KeyError:
                pass

            try:
                json_item = material_dict[argument]
                material_id = dict([("@id", json_item["@id"])])
                json_list.append(material_id)
            except KeyError:
                pass

            try:
                json_item = data_dict[argument]
                data_id = dict([("@id", json_item["@id"])])
                json_list.append(data_id)
            except KeyError:
                pass
        return json_list

    def createExecuteStudyProtocol(self, process_node_name, process_node):
        json_item = dict([
                   ("@id", self.getIdentifier(
                       "protocol", process_node.protocol))
                ])
        return json_item

    def createStudyAssaysList(self, assays, sample_dict):
        json_list = []
        for assay in assays:
            characteristics_categories_list = \
                self.createCharacteristicsCategories(assay.nodes)
            unit_categories_list = self.createUnitsCategories(assay.nodes)
            source_dict = self.createSourcesDictionary(assay.nodes)
            sample_list = self.createSampleReferenceDict(
                assay.nodes, sample_dict)
            material_dict = self.createMaterialDictionary(assay.nodes)
            data_dict = self.createDataFiles(assay.nodes)
            assay_name = assay.metadata['Study Assay File Name']
            assay_identifier = self.generateIdentifier("assay", assay_name)
            json_item = dict([
                ("@id", assay_identifier),
                ("filename", assay.metadata['Study Assay File Name']),
                ("measurementType", self.createOntologyAnnotation(
                    assay.metadata['Study Assay Measurement Type'],
                    assay.metadata[
                        'Study Assay Measurement Type Term Source REF'],
                    assay.metadata['Study Assay Measurement Type '
                                   'Term Accession Number'])),
                ("technologyType", self.createOntologyAnnotation(
                    assay.metadata['Study Assay Technology Type'],
                    assay.metadata['Study Assay Technology Type '
                                   'Term Source REF'],
                    assay.metadata['Study Assay Technology Type '
                                   'Term Accession Number'])),
                ("technologyPlatform", assay.metadata[
                    'Study Assay Technology Platform']),
                ("characteristicCategories", characteristics_categories_list),
                ("unitCategories", unit_categories_list),
                ("materials", dict([
                    ("samples", sample_list),
                    ("otherMaterials", list(material_dict.values()))
                ])),
                ("dataFiles", list(data_dict.values())),
                ("processSequence", self.createProcessSequence(
                    assay.process_nodes, source_dict, sample_dict,
                    material_dict, data_dict))
                ])
            json_list.append(json_item)
        return json_list

    def createFromNodeComments(self, node):
        comments = []
        for key in [key for key in node.metadata.keys() if
                    _RX_COMMENTS.match(key)]:
            comments.append(self.createComment(
                _RX_COMMENTS.findall(key)[0], getattr(
                    node.metadata[key][0],
                    _RX_COMMENTS.findall(key)[0].replace(' ', '_'))))
        return comments

    def createDataFiles(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype.endswith(" File"):
                data_identifier = self.generateIdentifier("data", node_index)
                json_item = dict([
                    ("@id", data_identifier),
                    ("name", nodes[node_index].name),
                    ("type", nodes[node_index].ntype),
                    ("comments", self.createFromNodeComments(
                        nodes[node_index]))
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createSampleDictionary(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype == "Sample Name":
                sample_identifier = self.generateIdentifier(
                    "sample", node_index)

                json_item = dict([
                        ("@id", sample_identifier),
                        ("name", node_index),
                        ("factorValues", self.createValueList(
                            self.FACTOR_VALUE, node_index, nodes[node_index])),
                        ("characteristics", self.createValueList(
                            self.CHARACTERISTICS, node_index,
                            nodes[node_index]))
                    ])

                #  derivesFrom sources
                try:
                    json_list = []
                    for source_name in nodes[node_index].derivesFrom:
                        source_index = "source-"+source_name
                        source_identifier = self.getIdentifier(
                            "source", source_index)
                        json_list.append(dict([("@id", source_identifier)]))

                    json_item["derivesFrom"] = json_list
                except KeyError:
                    log.error("There is no source declared for sample %s",
                              node_index)

                json_dict.update({node_index: json_item})

        return json_dict

    def createSampleReferenceDict(self, nodes, sample_dict):
        json_dict = []
        for node_index in nodes:
            node = nodes[node_index]
            if node.ntype == "Sample Name":
                sample_identifier = self.getIdentifier("sample", node_index)
            if sample_identifier:
                json_dict.append(dict([("@id", sample_identifier)]))
            else:
                log.warning(
                    "Warning: sample identifier has not been defined "
                    "before %s", node_index)
            try:
                sample_json = sample_dict[node_index]
                new_characteristics = self.createValueList(
                    self.CHARACTERISTICS, node_index, node)
                sample_json["characteristics"] = \
                    sample_json["characteristics"] + new_characteristics
                sample_dict[node_index] = sample_json
            except KeyError:
                log.warning(
                    "Warning: the sample %s has not been defined at the "
                    "study level", node_index)

        return json_dict

    def createSourcesDictionary(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype == "Source Name":
                source_identifier = self.generateIdentifier(
                    "source", node_index)
                json_item = dict([
                    ("@id", source_identifier),
                    ("name", node_index),
                    ("characteristics", self.createValueList(
                        self.CHARACTERISTICS, node_index, nodes[node_index])),
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createMaterialDictionary(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype != "Source Name" and \
                    nodes[node_index].ntype != "Sample Name" and \
                    nodes[node_index].ntype.find("File") == -1:
                material_identifier = self.generateIdentifier(
                    "material", node_index)
                json_item = dict([
                    ("@id", material_identifier),
                    ("name", node_index),
                    ("type", nodes[node_index].ntype),
                    ("characteristics", self.createValueList(
                        self.CHARACTERISTICS, node_index, nodes[node_index])),
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createCharacteristicsCategories(self, nodes):
        json_list = []
        for node_index in nodes:
            node = nodes[node_index]
            for header in node.metadata:
                if (not header.startswith(self.CHARACTERISTICS)) and \
                    (not header == self.MATERIAL_TYPE) and \
                        (not header == self.LABEL):
                    continue
                value_header = header.replace("]", "").split("[")[-1]
                if header == self.MATERIAL_TYPE:
                    value_header = self.MATERIAL_TYPE
                if header == self.LABEL:
                    value_header = self.LABEL

                characteristic_category_identifier = self.getIdentifier(
                    self.CHARACTERISTIC_CATEGORY, value_header)
                if characteristic_category_identifier:
                    continue

                characteristic_category_identifier = self.generateIdentifier(
                    self.CHARACTERISTIC_CATEGORY, value_header)

                json_item = dict([])
                if value_header.startswith("http"):
                    #  the header has an ontology annotation
                    #  TODO - get a test dataset
                    pass
                else:
                    json_item = dict([
                        ("@id", characteristic_category_identifier),
                        ("characteristicType",
                         self.createOntologyAnnotation(value_header, "", ""))
                    ])

                json_list.append(json_item)
        return json_list

    def createUnitsCategories(self, nodes):
        json_list = []
        for node_index in nodes:
            node = nodes[node_index]
            for header in node.metadata:
                if not header.startswith(self.CHARACTERISTICS) and \
                        not header.startswith(self.FACTOR_VALUE) and \
                        not header.startswith(self.PARAMETER_VALUE):
                    continue
                value_attributes = node.metadata[header][0]
                try:
                    unit = value_attributes.Unit
                except AttributeError:
                    continue

                unit_category_identifier = self.getIdentifier(self.UNIT, unit)
                if unit_category_identifier:
                    continue

                unit_category_identifier = self.generateIdentifier(
                    self.UNIT, value_attributes.Unit)
                json_item = dict([
                        ("@id", unit_category_identifier),
                    ])
                json_item.update(self.createOntologyAnnotation(
                    value_attributes.Unit, value_attributes.Term_Source_REF,
                    value_attributes.Term_Accession_Number))
                json_list.append(json_item)
        return json_list

    @staticmethod
    def convert_num(s):
        try:
            return int(s)
        except ValueError:
            try:
                return float(s)
            except ValueError:
                return s

    def createValueList(self, column_name, node_name, node):
        """
            Method for the creation of factor, characteristics and
            parameter values
        """
        json_list = []
        for header in node.metadata:
            if header.startswith(column_name) or \
                header == self.MATERIAL_TYPE or header == self.LABEL or \
                    header == self.ARRAY_DESIGN_REF:
                value_header = header.replace("]", "").split("[")[-1]

                value_attributes = node.metadata[header][0]
                value = self.convert_num(value_attributes[0])
                header_type = None

                if column_name.strip() == self.CHARACTERISTICS:
                    if header not in node.attributes:
                        continue
                    if header == self.MATERIAL_TYPE:
                        value_header = self.MATERIAL_TYPE
                    elif header == self.LABEL:
                        value_header = self.LABEL
                    header_type = self.CHARACTERISTIC_CATEGORY
                elif column_name.strip() == self.FACTOR_VALUE:
                    if header not in node.attributes:
                        continue
                    header_type = "factor"
                elif column_name.strip() == self.PARAMETER_VALUE:
                    if header not in node.parameters:
                        continue
                    if header == self.ARRAY_DESIGN_REF:
                        value_header = self.ARRAY_DESIGN_REF
                    header_type = "parameter"

                category_identifier = self.getIdentifier(
                    header_type, value_header)

                if value_header is None or category_identifier is None:
                    try:
                        unit_identifier = self.getIdentifier(
                            self.UNIT, value_attributes.Unit)
                        value_json = dict([
                            ("value", value),
                            ("unit", dict([("@id", unit_identifier)]))])
                        json_list.append(value_json)
                        continue
                    except AttributeError:
                        try:
                            value_json = dict([
                                 ("value", self.createOntologyAnnotation(
                                     value, value_attributes.Term_Source_REF,
                                     value_attributes.Term_Accession_Number))
                                 ])
                            json_list.append(value_json)
                            continue
                        except AttributeError:
                            value_json = dict([("value", value)])
                            json_list.append(value_json)
                else:
                    try:
                        unit_identifier = self.getIdentifier(
                            self.UNIT, value_attributes.Unit)
                        value_json = dict([
                            ("category", dict([("@id", category_identifier)])),
                            ("value", value),
                            ("unit", dict([("@id", unit_identifier)]))
                            ])
                        json_list.append(value_json)
                        continue
                    except AttributeError:
                        try:
                            value_json = dict([("category", dict([
                                 ("@id", category_identifier)])),
                             ("value", self.createOntologyAnnotation(
                                 value, value_attributes.Term_Source_REF,
                                 value_attributes.Term_Accession_Number))
                             ])
                            json_list.append(value_json)
                            continue
                        except AttributeError:
                            value_json = dict([
                                ("category", dict([
                                    ("@id", category_identifier)])),
                                ("value", value)])
                            json_list.append(value_json)
        return json_list



# -*- coding: utf-8 -*-
"""Functions for reading and writing SampleTab.

Functions for reading and writing SampleTab. SampleTab content is loaded into
an in-memory representation using the ISA Data Model implemented in the
isatools.model package.
"""
import io
import logging
from io import StringIO

import numpy as np
import pandas as pd

from progressbar import ETA, Bar, ProgressBar, SimpleProgress

from isatools import logging as isa_logging
from isatools.model import (
    Characteristic,
    Comment,
    FactorValue,
    Investigation,
    OntologyAnnotation,
    OntologySource,
    Person,
    Process,
    Protocol,
    Sample,
    Source,
    Study,
    StudyFactor,
)


log = logging.getLogger('isatools')


def _peek(f):
    """Peek at the next line without moving to the next line. This function
    get the position of the next line, reads the next line, then resets the
    file pointer to the original position

    :param f: A file-like buffer object
    :return: The next line past the current line
    """
    position = f.tell()
    line = f.readline()
    f.seek(position)
    return line


def _read_tab_section(f, sec_key, next_sec_key=None):
    """Slices a file by section delimited by section keys

    :param f: A file-like buffer object
    :param sec_key: Delimiter key of beginning of section
    :param next_sec_key: Delimiter key of end of section
    :return: A memory file of the section slice, as a string buffer object
    """
    line = f.readline()
    normed_line = line.rstrip()
    if normed_line[0] == '"':
        normed_line = normed_line[1:]
    if normed_line[len(normed_line) - 1] == '"':
        normed_line = normed_line[:len(normed_line) - 1]
    if not normed_line == sec_key:
        raise IOError("Expected: " + sec_key + " section, but got: "
                      + normed_line)
    memf = io.StringIO()
    while not _peek(f=f).rstrip() == next_sec_key:
        line = f.readline()
        if not line:
            break
        memf.write(line.rstrip() + '\n')
    memf.seek(0)
    return memf


def read_sampletab_msi(fp):
    """Read in a SampleTab and return a DataFrame of the MSI section

    :param fp: A file-like buffer object pointing to the sample tab file
    :return: A DataFrame with the MSI section
    """

    def _build_msi_df(f):
        """Builds the DataFrame from the MSI buffer

        :param f: A file-like buffer of the MSI section
        :return: DataFrame of the MSI section
        """
        f = strip_comments(f)
        df = pd.read_csv(f, names=range(0, 128), sep='\t', engine='python',
                         encoding='utf-8').dropna(axis=1, how='all')
        # load MSI section
        df = df.T  # transpose MSI section
        df.replace(np.nan, '', regex=True, inplace=True)
        # Strip out the nan entries
        df.reset_index(inplace=True)
        # Reset index so it is accessible as column
        df.columns = df.iloc[0]
        # If all was OK, promote this row to the column headers
        df = df.reindex(df.index.drop(0))
        # Reindex the DataFrame
        return df

    # Read in MSI section into DataFrames first
    msi_df = _build_msi_df(_read_tab_section(
        f=fp,
        sec_key='[MSI]',
        next_sec_key='[SCD]'
    ))
    return msi_df


def get_value(object_column, column_group, object_series,
              ontology_source_map, unit_categories):
    """Gets the appropriate value for a give column group

   :param object_column: The object's column header name, e.g. Sample Name
   :param column_group: The column group that includes the object's qualifiers
   :param object_series: Pandas DataFrame Series for the row
   :param ontology_source_map: A mapping to the OntologySource objects
   created after parsing the investigation file
   :param unit_categories: A map of unit categories to reference
   :return: The appropriate value and unit according to the columns parsed,
   e.g. (str, None) (float, Unit), (OntologyAnnotation, None)
   """
    cell_value = object_series[object_column]

    column_index = list(column_group).index(object_column)

    try:
        offset_1r_col = column_group[column_index + 1]
        offset_2r_col = column_group[column_index + 2]
    except IndexError:
        return cell_value, None

    if offset_1r_col.startswith('Term Source REF') \
            and offset_2r_col.startswith('Term Source ID'):

        value = OntologyAnnotation(term=str(cell_value))

        term_source_value = object_series[offset_1r_col]

        if term_source_value != '':

            try:
                value.term_source = ontology_source_map[term_source_value]
            except KeyError:
                log.warning('term source: ', term_source_value, ' not found')

        term_accession_value = str(object_series[offset_2r_col])

        if term_accession_value != '':
            value.term_accession = term_accession_value

        return value, None

    try:
        offset_3r_col = column_group[column_index + 3]
    except IndexError:
        return cell_value, None

    if offset_1r_col.startswith('Unit') and \
            offset_2r_col.startswith('Term Source REF') \
            and offset_3r_col.startswith('Term Source ID'):

        category_key = object_series[offset_1r_col]

        try:
            unit_term_value = unit_categories[category_key]
        except KeyError:
            unit_term_value = OntologyAnnotation(term=category_key)
            unit_categories[category_key] = unit_term_value

            unit_term_source_value = object_series[offset_2r_col]

            if unit_term_source_value != '':

                try:
                    unit_term_value.term_source = \
                        ontology_source_map[unit_term_source_value]
                except KeyError:
                    log.warning('term source: ', unit_term_source_value,
                                ' not found')

            term_accession_value = object_series[offset_3r_col]

            if term_accession_value != '':
                unit_term_value.term_accession = term_accession_value

        return cell_value, unit_term_value

    else:
        return cell_value, None


def load(FP):
    """Load a SampleTab into ISA Data Model objects

    :param FP: A file-like buffer object pointing to the SampleTab
    :return: An Investigation object
    """

    msi_df = read_sampletab_msi(FP)

    ISA = Investigation()

    for _, row in msi_df[["Term Source Name", "Term Source URI",
                          "Term Source Version"]]\
            .replace('', np.nan).dropna(axis=0, how='all').iterrows():
        ontology_source = OntologySource(name=row["Term Source Name"],
                                         file=row["Term Source URI"],
                                         version=row["Term Source Version"],
                                         description=row["Term Source Name"])
        ISA.ontology_source_references.append(ontology_source)

    row = msi_df[["Submission Title", "Submission Identifier",
                  "Submission Description", "Submission Version",
                  "Submission Reference Layer", "Submission Release Date",
                  "Submission Update Date"]].iloc[0]
    ISA.identifier = row["Submission Identifier"]
    ISA.title = row["Submission Title"]
    ISA.descriptiondescription = row["Submission Description"]
    ISA.submission_date = row["Submission Release Date"]
    ISA.comments = [
        Comment(name="Submission Version", value=row["Submission Version"]),
        Comment(name="Submission Reference Layer",
                value=row["Submission Reference Layer"]),
        Comment(
            name="Submission Update Date",
            value=row["Submission Update Date"]),
    ]

    try:
        for _, row in msi_df[["Person Last Name", "Person First Name",
                              "Person Initials", "Person Email",
                              "Person Role"]].replace(
                '', np.nan).dropna(axis=0, how='all').iterrows():
            person = Person(last_name=row['Person Last Name'],
                            first_name=row['Person First Name'],
                            mid_initials=row['Person Initials'],
                            email=row['Person Email'],
                            roles=[OntologyAnnotation(row['Person Role'])])
            ISA.contacts.append(person)
    except KeyError:
        pass
        # skip if no person part of MSI section is present, as in GSB-3.txt

    for i, row in msi_df[
        ["Organization Name", "Organization Address", "Organization URI",
         "Organization Email",
         "Organization Role"]].replace('', np.nan).dropna(
            axis=0, how='all').iterrows():
        ISA.comments.extend([
            Comment(name="Organization Name.{}".format(i),
                    value=row["Organization Name"]),
            Comment(name="Organization Address.{}".format(i),
                    value=row["Organization Address"]),
            Comment(name="Organization URI.{}".format(i),
                    value=row["Organization URI"]),
            Comment(name="Organization Email.{}".format(i),
                    value=row["Organization Email"]),
            Comment(name="Organization Role.{}".format(i),
                    value=row["Organization Role"])
        ])

    # Read in SCD section into DataFrame first
    FP = strip_comments(FP)
    scd_df = pd.read_csv(_read_tab_section(f=FP, sec_key='[SCD]'),
                         sep='\t', encoding='utf-8').fillna('')

    study = Study(filename="s_{}.txt".format(ISA.identifier))
    study.protocols = [Protocol(
        name='sample collection',
        protocol_type=OntologyAnnotation(term='sample collection'))]
    protocol_map = {
        "sample collection": Protocol(
            name="sample collection",
            protocol_type=OntologyAnnotation(term="sample collection"))
    }
    study.protocols = list(protocol_map.values())
    study.factors = [
        StudyFactor(name="Group Name"),
        StudyFactor(name="Group Accession")
    ]
    sources, samples, processes, characteristic_categories, unit_categories = \
        GenericSampleTabProcessSequenceFactory(
            ontology_sources=ISA.ontology_source_references,
            study_factors=study.factors).create_from_df(scd_df)
    study.sources = list(sources.values())
    study.samples = list(samples.values())
    study.process_sequence = list(processes.values())
    study.characteristic_categories = list(characteristic_categories.values())
    study.units = list(unit_categories.values())
    for process in study.process_sequence:
        try:
            process.executes_protocol = protocol_map[process.executes_protocol]
        except KeyError:
            try:
                unknown_protocol = protocol_map['unknown']
            except KeyError:
                protocol_map['unknown'] = Protocol(
                    name="unknown protocol",
                    description="This protocol was auto-generated where a "
                                "protocol could not be determined.")
                unknown_protocol = protocol_map['unknown']
                study.protocols.append(unknown_protocol)
            process.executes_protocol = unknown_protocol
    ISA.studies = [study]
    return ISA


class GenericSampleTabProcessSequenceFactory:
    """The GenericSampleTabProcessSequenceFactory is used to parse the
    tables and build the process sequences representing the
    experimental graphs"""

    def __init__(self, ontology_sources=None, study_factors=None):
        self.ontology_sources = ontology_sources
        self.factors = study_factors

    def create_from_df(self, DF):
        """Create the process sequences from the table DataFrame

        :param DF: Table DataFrame
        :return: List of Processes coressponding to the process sequences. The
        Processes are linked appropriately to all other ISA content objects,
        such as Samples, DataFiles, and to each other.
        """
        if self.ontology_sources is not None:
            ontology_source_map = dict(map(lambda x: (x.name, x),
                                           self.ontology_sources))
        else:
            ontology_source_map = {}

        samples = {}
        characteristic_categories = {}
        unit_categories = {}
        processes = {}

        try:
            cproject = DF["Characteristic[project]"].drop_duplicates()
            if len(cproject.index) == 1:
                log.info("{} project type".format(cproject.iloc[0]))
        except KeyError:
            log.info("Assuming default project type")

        try:
            samples.update(dict(map(lambda x: (x, Source(
                comments=[Comment(name="Sample Accession", value=x)])),
                DF["Sample Accession"].loc[DF["Derived From"] == ""]
                .drop_duplicates())))
        except KeyError:
            pass

        try:
            samples.update(dict(map(lambda x: (x, Sample(
                comments=[Comment(name="Sample Accession", value=x)])),
                DF["Sample Accession"].loc[DF["Derived From"] != ""]
                .drop_duplicates())))
        except KeyError:
            pass

        for sample_key in samples.keys():
            sample = samples[sample_key]

            row = DF[DF["Sample Accession"] == sample_key].iloc[0]
            # there should only be one row with accession
            sample.name = row["Sample Name"]

            if row["Sample Accession"] != "":

                try:
                    category = characteristic_categories["Sample Accession"]
                except KeyError:
                    category = OntologyAnnotation(term="Sample Accession")
                    characteristic_categories["Sample Accession"] = category
                sample.characteristics.append(Characteristic(
                    category=category, value=row["Sample Accession"]))

            if row["Sample Description"] != "":
                try:
                    category = characteristic_categories["Sample Description"]
                except KeyError:
                    category = OntologyAnnotation(term="Sample Description")
                    characteristic_categories["Sample Description"] = category
                sample.characteristics.append(Characteristic(
                    category=category, value=row["Sample Description"]))

            if row["Derived From"] != "":
                try:
                    category = characteristic_categories["Derived From"]
                except KeyError:
                    category = OntologyAnnotation(term="Derived From")
                    characteristic_categories["Derived From"] = category
                sample.characteristics.append(Characteristic(
                    category=category, value=row["Derived From"]))

            try:
                if row["Child Of"] != "":
                    try:
                        category = characteristic_categories["Child Of"]
                    except KeyError:
                        category = OntologyAnnotation(term="Child Of")
                        characteristic_categories["Child Of"] = category
                    sample.characteristics.append(Characteristic(
                        category=category, value=row["Child Of"]))
            except KeyError:
                pass  # skip if Child Of is not present in sample table

            if row["Group Name"] != "":
                if isinstance(sample, Sample):
                    factor_hits = [f for f in self.factors
                                   if f.name == "Group Name"]
                    if len(factor_hits) == 1:
                        factor = factor_hits[0]
                    else:
                        raise ValueError("Could not resolve Study Factor from "
                                         "Group Name")
                    fv = FactorValue(factor_name=factor)
                    v = row["Group Name"]
                    fv.value = v
                    sample.factor_values.append(fv)
                else:
                    category_key = "Group Name"
                    try:
                        category = characteristic_categories[category_key]
                    except KeyError:
                        category = OntologyAnnotation(term=category_key)
                        characteristic_categories[category_key] = category
                    characteristic = Characteristic(category=category)
                    v = row["Group Name"]
                    characteristic.value = v

            if row["Group Accession"] != "":
                if isinstance(sample, Sample):
                    factor_hits = [f for f in self.factors
                                   if f.name == "Group Accession"]
                    if len(factor_hits) == 1:
                        factor = factor_hits[0]
                    else:
                        raise ValueError("Could not resolve Study Factor "
                                         "from Group Accession")
                    fv = FactorValue(factor_name=factor)
                    v = row["Group Accession"]
                    fv.value = v
                    sample.factor_values.append(fv)
                else:
                    category_key = "Group Accession"
                    try:
                        category = characteristic_categories[category_key]
                    except KeyError:
                        category = OntologyAnnotation(term=category_key)
                        characteristic_categories[category_key] = category
                    characteristic = Characteristic(category=category)
                    v = row["Group Accession"]
                    characteristic.value = v

            for col in [x for x in DF.columns if x.startswith(
                    "Characteristic[")]:  # build object map
                category_key = col[15:col.rfind("]")]
                try:
                    category = characteristic_categories[category_key]
                except KeyError:
                    category = OntologyAnnotation(term=category_key)
                    characteristic_categories[category_key] = category

                characteristic = Characteristic(category=category)

                v, u = get_value(col, DF.columns, row,
                                 ontology_source_map, unit_categories)

                characteristic.value = v
                characteristic.unit = u

                sample.characteristics.append(characteristic)

            sample_accession = row["Derived From"]

            try:
                source = samples[sample_accession]
                sample.derives_from.append(source)
            except KeyError:
                pass

        sample_collection_protocol = "sample collection"

        for _, row in DF[["Sample Accession", "Derived From"]].iterrows():
            sample_accession = row["Sample Accession"]
            sample = samples[sample_accession]
            derived_from_accession = row["Derived From"]
            if derived_from_accession == "":
                continue
            derived_from_sample = samples[derived_from_accession]
            sample.derived_from = derived_from_sample
            process_key = ":".join([derived_from_accession,
                                    sample_collection_protocol])
            try:
                process = processes[process_key]
            except KeyError:
                process = Process(executes_protocol=sample_collection_protocol)
                processes[process_key] = process
            if derived_from_sample not in process.inputs:
                process.inputs.append(derived_from_sample)
            if sample not in process.outputs:
                process.outputs.append(sample)

        sources = dict([x for x in samples.items()
                        if isinstance(x[1], Source)])
        study_samples = dict([x for x in samples.items()
                              if isinstance(x[1], Sample)])
        return sources, study_samples, processes, characteristic_categories, \
            unit_categories


def dumps(investigation):
    """Dumps out an ISA Investigation to a SampleTab string

    :param investigation: An Investigation object
    :return: String of the SampleTab
    """

    # build MSI section

    metadata_DF = pd.DataFrame(columns=(
        "Submission Title",
        "Submission Identifier",
        "Submission Description",
                               "Submission Version",
                               "Submission Reference Layer",
                               "Submission Release Date",
                               "Submission Update Date"))
    iversion_hits = [x for x in investigation.comments
                     if x.name == "Submission Version"]
    if len(iversion_hits) == 1:
        investigation_version = iversion_hits[0].value
    else:
        investigation_version = ""
    ireference_layer_hits = [x for x in investigation.comments
                             if x.name == "Submission Reference Layer"]
    if len(ireference_layer_hits) == 1:
        investigation_reference_layer = ireference_layer_hits[0].value
    else:
        investigation_reference_layer = ""
    iversion_update_date = [x for x in investigation.comments
                            if x.name == "Submission Update Date"]
    if len(iversion_update_date) == 1:
        investigation_update_date = iversion_update_date[0].value
    else:
        investigation_update_date = ""
    metadata_DF.loc[0] = [
        investigation.title,
        investigation.identifier,
        investigation.description,
        investigation_version,
        investigation_reference_layer,
        investigation.submission_date,
        investigation_update_date
    ]

    org_DF = pd.DataFrame(columns=(
        "Organization Name", "Organization Address", "Organization URI",
        "Organization Email", "Organization Role"))
    org_name_hits = [x for x in investigation.comments
                     if x.name.startswith("Organization Name")]
    org_address_hits = [x for x in investigation.comments
                        if x.name.startswith("Organization Address")]
    org_uri_hits = [x for x in investigation.comments
                    if x.name.startswith("Organization URI")]
    org_email_hits = [x for x in investigation.comments
                      if x.name.startswith("Organization Email")]
    org_role_hits = [x for x in investigation.comments
                     if x.name.startswith("Organization Role")]
    for i, org_name in enumerate(org_name_hits):
        try:
            org_name = org_name_hits[i].value
        except IndexError:
            org_name = ""
        try:
            org_address = org_address_hits[i].value
        except IndexError:
            org_address = ""
        try:
            org_uri = org_uri_hits[i].value
        except IndexError:
            org_uri = ""
        try:
            org_email = org_email_hits[i].value
        except IndexError:
            org_email = ""
        try:
            org_role = org_role_hits[i].value
        except IndexError:
            org_role = ""
        org_DF.loc[i] = [
            org_name,
            org_address,
            org_uri,
            org_email,
            org_role
        ]

    people_DF = pd.DataFrame(columns=(
        "Person Last Name", "Person Initials", "Person First Name",
        "Person Email", "Person Role"))
    for i, contact in enumerate(investigation.contacts):
        if len(contact.roles) == 1:
            role = contact.roles[0].term
        else:
            role = ""
        people_DF.loc[i] = [
            contact.last_name,
            contact.mid_initials,
            contact.first_name,
            contact.email,
            role
        ]

    term_sources_DF = pd.DataFrame(columns=(
        "Term Source Name", "Term Source URI", "Term Source Version"))
    for i, term_source in enumerate(investigation.ontology_source_references):
        term_sources_DF.loc[i] = [
            term_source.name,
            term_source.file,
            term_source.version
        ]
    msi_DF = pd.concat(
        [metadata_DF, org_DF, people_DF, term_sources_DF], axis=1)
    msi_DF = msi_DF.set_index("Submission Title").T
    msi_DF = msi_DF.replace('', np.nan)
    msi_memf = StringIO()
    msi_DF.to_csv(
        path_or_buf=msi_memf,
        index=True,
        sep='\t',
        encoding='utf-8',
        index_label="Submission Title")
    msi_memf.seek(0)

    scd_DF = pd.DataFrame(columns=(
        "Sample Name", "Sample Accession", "Sample Description",
        "Derived From", "Group Name", "Group Accession"))

    all_samples = []
    for study in investigation.studies:
        all_samples += study.sources
        all_samples += study.samples

    all_samples = list(set(all_samples))
    if isa_logging.show_pbars:
        pbar = ProgressBar(min_value=0, max_value=len(all_samples),
                           widgets=['Writing {} samples: '.format(
                               len(all_samples)), SimpleProgress(),
            Bar(left=" |", right="| "), ETA()]).start()
    else:
        def pbar(x): return x
    for i, s in pbar(enumerate(all_samples)):
        derived_from = ""
        if isinstance(s, Sample) and s.derives_from is not None:
            if len(s.derives_from) == 1:
                derived_from_obj = s.derives_from[0]
                derives_from_accession_hits = [
                    x for x in derived_from_obj.characteristics
                    if x.category.term == "Sample Accession"]
                if len(derives_from_accession_hits) == 1:
                    derived_from = derives_from_accession_hits[0].value
                else:
                    log.warning(
                        "WARNING! No Sample Accession available so "
                        "referencing Derived From relation using "
                        "Sample Name \"{}\" instead".format(
                            derived_from_obj.name))
                    derived_from = derived_from_obj.name
        sample_accession_hits = [x for x in s.characteristics
                                 if x.category.term == "Sample Accession"]
        if len(sample_accession_hits) == 1:
            sample_accession = sample_accession_hits[0].value
        else:
            sample_accession = ""
        sample_description_hits = [x for x in s.characteristics
                                   if x.category.term == "Sample Description"]
        if len(sample_description_hits) == 1:
            sample_description = sample_description_hits[0].value
        else:
            sample_description = ""

        if isinstance(s, Sample):
            group_name_hits = [x for x in s.factor_values
                               if x.factor_name.name == "Group Name"]
            if len(group_name_hits) == 1:
                group_name = group_name_hits[0].value
            else:
                group_name = ""
            group_accession_hits = [x for x in s.factor_values
                                    if x.factor_name.name == "Group Accession"]
            if len(group_accession_hits) == 1:
                group_accession = group_accession_hits[0].value
            else:
                group_accession = ""
        else:
            group_name_hits = [x for x in s.characteristics
                               if x.category.term == "Group Name"]
            if len(group_name_hits) == 1:
                group_name = group_name_hits[0].value
            else:
                group_name = ""
            group_accession_hits = [x for x in s.characteristics
                                    if x.category.term == "Group Accession"]
            if len(group_accession_hits) == 1:
                group_accession = group_accession_hits[0].value
            else:
                group_accession = ""

        scd_DF.loc[i, "Sample Name"] = s.name
        scd_DF.loc[i, "Sample Accession"] = sample_accession
        scd_DF.loc[i, "Sample Description"] = sample_description
        scd_DF.loc[i, "Derived From"] = derived_from
        scd_DF.loc[i, "Group Name"] = group_name
        scd_DF.loc[i, "Group Accession"] = group_accession

        characteristics = [
            x for x in s.characteristics if x.category.term not in [
                "Sample Description",
                "Derived From",
                "Sample Accession"]]
        for characteristic in characteristics:
            characteristic_label = "Characteristic[{}]".format(
                characteristic.category.term)
            if characteristic_label not in scd_DF.columns:
                scd_DF[characteristic_label] = ""
                for val_col in get_value_columns(
                        characteristic_label, characteristic):
                    scd_DF[val_col] = ""
            if isinstance(characteristic.value, (int, float)
                          ) and characteristic.unit:
                if isinstance(characteristic.unit, OntologyAnnotation):
                    scd_DF.loc[i, characteristic_label] = characteristic.value
                    scd_DF.loc[i, characteristic_label +
                               ".Unit"] = characteristic.unit.term
                    scd_DF.loc[i, characteristic_label +
                               ".Unit.Term Source REF"]\
                        = characteristic.unit.term_source.name \
                        if characteristic.unit.term_source else ""
                    scd_DF.loc[i, characteristic_label +
                               ".Unit.Term Accession Number"] = \
                        characteristic.unit.term_accession
                else:
                    scd_DF.loc[i, characteristic_label] = characteristic.value
                    scd_DF.loc[i, characteristic_label +
                               ".Unit"] = characteristic.unit
            elif isinstance(characteristic.value, OntologyAnnotation):
                scd_DF.loc[i, characteristic_label] = characteristic.value.term
                scd_DF.loc[i, characteristic_label + ".Term Source REF"] = \
                    characteristic.value.term_source.name \
                    if characteristic.value.term_source else ""
                scd_DF.loc[i, characteristic_label
                           + ".Term Accession Number"] = \
                    characteristic.value.term_accession
            else:
                scd_DF.loc[i, characteristic_label] = characteristic.value

    scd_DF = scd_DF.replace('', np.nan)
    columns = list(scd_DF.columns)
    for i, col in enumerate(columns):
        if col.endswith("Term Source REF"):
            columns[i] = "Term Source REF"
        elif col.endswith("Term Accession Number"):
            columns[i] = "Term Source ID"
        elif col.endswith("Unit"):
            columns[i] = "Unit"
    scd_DF.columns = columns
    scd_memf = StringIO()
    scd_DF.to_csv(
        path_or_buf=scd_memf,
        index=False,
        sep='\t',
        encoding='utf-8')
    scd_memf.seek(0)

    sampletab_memf = StringIO()
    sampletab_memf.write("[MSI]\n")
    for line in msi_memf:
        sampletab_memf.write(line.rstrip() + '\n')
    sampletab_memf.write("[SCD]\n")
    for line in scd_memf:
        sampletab_memf.write(line.rstrip() + '\n')
    sampletab_memf.seek(0)

    return sampletab_memf.read()


def dump(investigation, out_fp):
    """Dump out an Investigation to a SampleTab file

    :param investigation: An Investigation file
    :param out_fp: A file-like buffer object to write to
    :return: None
    """
    sampletab_str = dumps(investigation)
    out_fp.write(sampletab_str)


def get_value_columns(label, x):
    """Generates the appropriate columns based on the value of the object.
    For example, if the object's .value value is an OntologyAnnotation,
    the ISA-Tab requires extra columns Term Source REF and
    Term Accession Number

    :param label: Header label needed for the object type, e.g. "Sample Name"
    :param x: The object of interest, e.g. a Sample() object
    :return: List of column labels, e.g. ["Sample Name.Term Source REF",
    "Sample Name.Term Accession Number"]
    """
    if isinstance(x.value, (int, float)) and x.unit:
        if isinstance(x.unit, OntologyAnnotation):
            return map(lambda x: "{0}.{1}".format(label, x),
                       ["Unit", "Unit.Term Source REF",
                        "Unit.Term Accession Number"])
        else:
            return ["{0}.Unit".format(label)]
    elif isinstance(x.value, OntologyAnnotation):
        return map(lambda x: "{0}.{1}".format(label, x), [
                   "Term Source REF", "Term Accession Number"])
    else:
        return []


def strip_comments(in_fp):
    """Strip out comment lines indicated by a # at start of line from a given
    file

    :param in_fp: A file-like buffer object
    :return: A memory file buffer object with comments stripped out
    """
    out_fp = StringIO()
    if not isinstance(in_fp, StringIO):
        out_fp.name = in_fp.name
    for line in in_fp.readlines():
        if line.lstrip().startswith('#'):
            pass
        else:
            out_fp.write(line)
    out_fp.seek(0)
    return out_fp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Generated Thu Jan 21 10:45:32 2016 by generateDS.py version 2.18a.
#
# Updated with 2to3 for Python 3 compatibility
#

import base64
import datetime as datetime_
import glob
import logging
import os
import re as re_
import sys
import warnings as warnings_

from lxml import etree as etree_


log = logging.getLogger('isatools')


def load(config_dir):
    config_dict = dict()
    for file in glob.iglob(os.path.join(config_dir, '*.xml')):
        try:
            config_obj = parse(inFileName=file, silence=True)
            measurement_type = config_obj.get_isatab_configuration()[0].get_measurement().get_term_label()
            technology_type = config_obj.get_isatab_configuration()[0].get_technology().get_term_label()
            config_dict[(measurement_type.lower(), technology_type.lower())] = config_obj
        except GDSParseError as parse_error:
            log.error(parse_error)
    return config_dict


def get_config(config_dict, measurement_type=None, technology_type=None):
    try:
        config = config_dict[(measurement_type, technology_type)].isatab_configuration[0]
        from collections import OrderedDict
        fields = dict()
        for field in config.field:
            fields[field.pos] = field
        for protocol_field in config.protocol_field:
            fields[protocol_field.pos] = protocol_field
        for structured_field in config.structured_field:
            fields[structured_field.pos] = structured_field
        sorted_fields = OrderedDict(sorted(fields.items(), key=lambda x: x[0]))
        sorted_config = list(sorted_fields.values())
    except KeyError:
        sorted_config = None
    return sorted_config


Validate_simpletypes_ = True


def parsexml_(infile, parser=None, **kwargs):
    if parser is None:
        # Use the lxml ElementTree compatible parser so that, e.g.,
        #   we ignore comments.
        parser = etree_.ETCompatXMLParser()
    doc = etree_.parse(infile, parser=parser, **kwargs)
    return doc

#
# User methods
#
# Calls to the methods in these classes are generated by generateDS.py.
# You can replace these methods by re-implementing the following class
#   in a module named generatedssuper.py.

try:
    from generatedssuper import GeneratedsSuper
except ImportError as exp:

    class GeneratedsSuper(object):
        tzoff_pattern = re_.compile(r'(\+|-)((0\d|1[0-3]):[0-5]\d|14:00)$')
        class _FixedOffsetTZ(datetime_.tzinfo):
            def __init__(self, offset, name):
                self.__offset = datetime_.timedelta(minutes=offset)
                self.__name = name
            def utcoffset(self, dt):
                return self.__offset
            def tzname(self, dt):
                return self.__name
            def dst(self, dt):
                return None
        def gds_format_string(self, input_data, input_name=''):
            return input_data
        def gds_validate_string(self, input_data, node=None, input_name=''):
            if not input_data:
                return ''
            else:
                return input_data
        def gds_format_base64(self, input_data, input_name=''):
            return base64.b64encode(input_data)
        def gds_validate_base64(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer(self, input_data, input_name=''):
            return '%d' % input_data
        def gds_validate_integer(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_integer_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    int(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of integers')
            return values
        def gds_format_float(self, input_data, input_name=''):
            return ('%.15f' % input_data).rstrip('0')
        def gds_validate_float(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_float_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_float_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of floats')
            return values
        def gds_format_double(self, input_data, input_name=''):
            return '%e' % input_data
        def gds_validate_double(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_double_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_double_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of doubles')
            return values
        def gds_format_boolean(self, input_data, input_name=''):
            return ('%s' % input_data).lower()
        def gds_validate_boolean(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_boolean_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_boolean_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                if value not in ('true', '1', 'false', '0', ):
                    raise_parse_error(
                        node,
                        'Requires sequence of booleans '
                        '("true", "1", "false", "0")')
            return values
        def gds_validate_datetime(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_datetime(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d.%s' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        @classmethod
        def gds_parse_datetime(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            time_parts = input_data.split('.')
            if len(time_parts) > 1:
                micro_seconds = int(float('0.' + time_parts[1]) * 1000000)
                input_data = '%s.%s' % (time_parts[0], micro_seconds, )
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt
        def gds_validate_date(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_date(self, input_data, input_name=''):
            _svalue = '%04d-%02d-%02d' % (
                input_data.year,
                input_data.month,
                input_data.day,
            )
            try:
                if input_data.tzinfo is not None:
                    tzoff = input_data.tzinfo.utcoffset(input_data)
                    if tzoff is not None:
                        total_seconds = tzoff.seconds + (86400 * tzoff.days)
                        if total_seconds == 0:
                            _svalue += 'Z'
                        else:
                            if total_seconds < 0:
                                _svalue += '-'
                                total_seconds *= -1
                            else:
                                _svalue += '+'
                            hours = total_seconds // 3600
                            minutes = (total_seconds - (hours * 3600)) // 60
                            _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            except AttributeError:
                pass
            return _svalue
        @classmethod
        def gds_parse_date(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            dt = datetime_.datetime.strptime(input_data, '%Y-%m-%d')
            dt = dt.replace(tzinfo=tz)
            return dt.date()
        def gds_validate_time(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_time(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%02d:%02d:%02d' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%02d:%02d:%02d.%s' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        def gds_validate_simple_patterns(self, patterns, target):
            # pat is a list of lists of strings/patterns.  We should:
            # - AND the outer elements
            # - OR the inner elements
            found1 = True
            for patterns1 in patterns:
                found2 = False
                for patterns2 in patterns1:
                    if re_.search(patterns2, target) is not None:
                        found2 = True
                        break
                if not found2:
                    found1 = False
                    break
            return found1
        @classmethod
        def gds_parse_time(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            if len(input_data.split('.')) > 1:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt.time()
        def gds_str_lower(self, instring):
            return instring.lower()
        def get_path_(self, node):
            path_list = []
            self.get_path_list_(node, path_list)
            path_list.reverse()
            path = '/'.join(path_list)
            return path
        Tag_strip_pattern_ = re_.compile(r'\{.*\}')
        def get_path_list_(self, node, path_list):
            if node is None:
                return
            tag = GeneratedsSuper.Tag_strip_pattern_.sub('', node.tag)
            if tag:
                path_list.append(tag)
            self.get_path_list_(node.getparent(), path_list)
        def get_class_obj_(self, node, default_class=None):
            class_obj1 = default_class
            if 'xsi' in node.nsmap:
                classname = node.get('{%s}type' % node.nsmap['xsi'])
                if classname is not None:
                    names = classname.split(':')
                    if len(names) == 2:
                        classname = names[1]
                    class_obj2 = globals().get(classname)
                    if class_obj2 is not None:
                        class_obj1 = class_obj2
            return class_obj1
        def gds_build_any(self, node, type_name=None):
            return None
        @classmethod
        def gds_reverse_node_mapping(cls, mapping):
            return dict(((v, k) for k, v in mapping.items()))


#
# If you have installed IPython you can uncomment and use the following.
# IPython is available from http://ipython.scipy.org/.
#

## from IPython.Shell import IPShellEmbed
## args = ''
## ipshell = IPShellEmbed(args,
##     banner = 'Dropping into IPython',
##     exit_msg = 'Leaving Interpreter, back to program.')

# Then use the following line where and when you want to drop into the
# IPython shell:
#    ipshell('<some message> -- Entering ipshell.\nHit Ctrl-D to exit')

#
# Globals
#

ExternalEncoding = 'ascii'
Tag_pattern_ = re_.compile(r'({.*})?(.*)')
String_cleanup_pat_ = re_.compile(r"[\n\r\s]+")
Namespace_extract_pat_ = re_.compile(r'{(.*)}(.*)')
CDATA_pattern_ = re_.compile(r"<!\[CDATA\[.*?\]\]>", re_.DOTALL)

#
# Support/utility functions.
#


def showIndent(outfile, level, pretty_print=True):
    if pretty_print:
        for idx in range(level):
            outfile.write('    ')


def quote_xml(inStr):
    "Escape markup chars, but do not modify CDATA sections."
    if not inStr:
        return ''
    s1 = (isinstance(inStr, str) and inStr or
          '%s' % inStr)
    s2 = ''
    pos = 0
    matchobjects = CDATA_pattern_.finditer(s1)
    for mo in matchobjects:
        s3 = s1[pos:mo.start()]
        s2 += quote_xml_aux(s3)
        s2 += s1[mo.start():mo.end()]
        pos = mo.end()
    s3 = s1[pos:]
    s2 += quote_xml_aux(s3)
    return s2


def quote_xml_aux(inStr):
    s1 = inStr.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    return s1


def quote_attrib(inStr):
    s1 = (isinstance(inStr, str) and inStr or
          '%s' % inStr)
    s1 = s1.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    if '"' in s1:
        if "'" in s1:
            s1 = '"%s"' % s1.replace('"', "&quot;")
        else:
            s1 = "'%s'" % s1
    else:
        s1 = '"%s"' % s1
    return s1


def quote_python(inStr):
    s1 = inStr
    if s1.find("'") == -1:
        if s1.find('\n') == -1:
            return "'%s'" % s1
        else:
            return "'''%s'''" % s1
    else:
        if s1.find('"') != -1:
            s1 = s1.replace('"', '\\"')
        if s1.find('\n') == -1:
            return '"%s"' % s1
        else:
            return '"""%s"""' % s1


def get_all_text_(node):
    if node.text is not None:
        text = node.text
    else:
        text = ''
    for child in node:
        if child.tail is not None:
            text += child.tail
    return text


def find_attr_value_(attr_name, node):
    attrs = node.attrib
    attr_parts = attr_name.split(':')
    value = None
    if len(attr_parts) == 1:
        value = attrs.get(attr_name)
    elif len(attr_parts) == 2:
        prefix, name = attr_parts
        namespace = node.nsmap.get(prefix)
        if namespace is not None:
            value = attrs.get('{%s}%s' % (namespace, name, ))
    return value


class GDSParseError(Exception):
    pass


def raise_parse_error(node, msg):
    msg = '%s (element %s/line %d)' % (msg, node.tag, node.sourceline, )
    raise GDSParseError(msg)


class MixedContainer:
    # Constants for category:
    CategoryNone = 0
    CategoryText = 1
    CategorySimple = 2
    CategoryComplex = 3
    # Constants for content_type:
    TypeNone = 0
    TypeText = 1
    TypeString = 2
    TypeInteger = 3
    TypeFloat = 4
    TypeDecimal = 5
    TypeDouble = 6
    TypeBoolean = 7
    TypeBase64 = 8
    def __init__(self, category, content_type, name, value):
        self.category = category
        self.content_type = content_type
        self.name = name
        self.value = value
    def getCategory(self):
        return self.category
    def getContenttype(self, content_type):
        return self.content_type
    def getValue(self):
        return self.value
    def getName(self):
        return self.name
    def export(self, outfile, level, name, namespace, pretty_print=True):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                outfile.write(self.value)
        elif self.category == MixedContainer.CategorySimple:
            self.exportSimple(outfile, level, name)
        else:    # category == MixedContainer.CategoryComplex
            self.value.export(outfile, level, namespace, name, pretty_print)
    def exportSimple(self, outfile, level, name):
        if self.content_type == MixedContainer.TypeString:
            outfile.write('<%s>%s</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeInteger or \
                self.content_type == MixedContainer.TypeBoolean:
            outfile.write('<%s>%d</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeFloat or \
                self.content_type == MixedContainer.TypeDecimal:
            outfile.write('<%s>%f</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeDouble:
            outfile.write('<%s>%g</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeBase64:
            outfile.write('<%s>%s</%s>' % (
                self.name, base64.b64encode(self.value), self.name))
    def to_etree(self, element):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                if len(element) > 0:
                    if element[-1].tail is None:
                        element[-1].tail = self.value
                    else:
                        element[-1].tail += self.value
                else:
                    if element.text is None:
                        element.text = self.value
                    else:
                        element.text += self.value
        elif self.category == MixedContainer.CategorySimple:
            subelement = etree_.SubElement(element, '%s' % self.name)
            subelement.text = self.to_etree_simple()
        else:    # category == MixedContainer.CategoryComplex
            self.value.to_etree(element)
    def to_etree_simple(self):
        if self.content_type == MixedContainer.TypeString:
            text = self.value
        elif (self.content_type == MixedContainer.TypeInteger or
                self.content_type == MixedContainer.TypeBoolean):
            text = '%d' % self.value
        elif (self.content_type == MixedContainer.TypeFloat or
                self.content_type == MixedContainer.TypeDecimal):
            text = '%f' % self.value
        elif self.content_type == MixedContainer.TypeDouble:
            text = '%g' % self.value
        elif self.content_type == MixedContainer.TypeBase64:
            text = '%s' % base64.b64encode(self.value)
        return text
    def exportLiteral(self, outfile, level, name):
        if self.category == MixedContainer.CategoryText:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        elif self.category == MixedContainer.CategorySimple:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        else:    # category == MixedContainer.CategoryComplex
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s",\n' % (
                    self.category, self.content_type, self.name,))
            self.value.exportLiteral(outfile, level + 1)
            showIndent(outfile, level)
            outfile.write(')\n')


class MemberSpec_(object):
    def __init__(self, name='', data_type='', container=0):
        self.name = name
        self.data_type = data_type
        self.container = container
    def set_name(self, name): self.name = name
    def get_name(self): return self.name
    def set_data_type(self, data_type): self.data_type = data_type
    def get_data_type_chain(self): return self.data_type
    def get_data_type(self):
        if isinstance(self.data_type, list):
            if len(self.data_type) > 0:
                return self.data_type[-1]
            else:
                return 'xs:string'
        else:
            return self.data_type
    def set_container(self, container): self.container = container
    def get_container(self): return self.container


def _cast(typ, value):
    if typ is None or value is None:
        return value
    return typ(value)

#
# Data representation classes.
#

class FieldType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, is_file_field=None, section=None, is_forced_ontology=None, header=None, data_type=None, is_multiple_value=None, is_hidden=None, is_required=None, description=None, default_value=None, value_format=None, list_values=None, generated_value_template=None, recommended_ontologies=None, value_range=None):
        self.original_tagname_ = None
        self.is_file_field = _cast(bool, is_file_field)
        self.section = _cast(None, section)
        self.is_forced_ontology = _cast(bool, is_forced_ontology)
        self.header = _cast(None, header)
        self.data_type = _cast(None, data_type)
        self.is_multiple_value = _cast(bool, is_multiple_value)
        self.is_hidden = _cast(bool, is_hidden)
        self.is_required = _cast(bool, is_required)
        self.description = description
        self.default_value = default_value
        self.value_format = value_format
        self.list_values = list_values
        self.generated_value_template = generated_value_template
        self.recommended_ontologies = recommended_ontologies
        self.value_range = value_range
    def factory(*args_, **kwargs_):
        if FieldType.subclass:
            return FieldType.subclass(*args_, **kwargs_)
        else:
            return FieldType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_description(self): return self.description
    def set_description(self, description): self.description = description
    def get_default_value(self): return self.default_value
    def set_default_value(self, default_value): self.default_value = default_value
    def get_value_format(self): return self.value_format
    def set_value_format(self, value_format): self.value_format = value_format
    def get_list_values(self): return self.list_values
    def set_list_values(self, list_values): self.list_values = list_values
    def get_generated_value_template(self): return self.generated_value_template
    def set_generated_value_template(self, generated_value_template): self.generated_value_template = generated_value_template
    def get_recommended_ontologies(self): return self.recommended_ontologies
    def set_recommended_ontologies(self, recommended_ontologies): self.recommended_ontologies = recommended_ontologies
    def get_value_range(self): return self.value_range
    def set_value_range(self, value_range): self.value_range = value_range
    def get_is_file_field(self): return self.is_file_field
    def set_is_file_field(self, is_file_field): self.is_file_field = is_file_field
    def get_section(self): return self.section
    def set_section(self, section): self.section = section
    def get_is_forced_ontology(self): return self.is_forced_ontology
    def set_is_forced_ontology(self, is_forced_ontology): self.is_forced_ontology = is_forced_ontology
    def get_header(self): return self.header
    def set_header(self, header): self.header = header
    def get_data_type(self): return self.data_type
    def set_data_type(self, data_type): self.data_type = data_type
    def get_is_multiple_value(self): return self.is_multiple_value
    def set_is_multiple_value(self, is_multiple_value): self.is_multiple_value = is_multiple_value
    def get_is_hidden(self): return self.is_hidden
    def set_is_hidden(self, is_hidden): self.is_hidden = is_hidden
    def get_is_required(self): return self.is_required
    def set_is_required(self, is_required): self.is_required = is_required
    def hasContent_(self):
        if (
            self.description is not None or
            self.default_value is not None or
            self.value_format is not None or
            self.list_values is not None or
            self.generated_value_template is not None or
            self.recommended_ontologies is not None or
            self.value_range is not None
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='FieldType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='FieldType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='FieldType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='FieldType'):
        if self.is_file_field is not None and 'is_file_field' not in already_processed:
            already_processed.add('is_file_field')
            outfile.write(' is-file-field="%s"' % self.gds_format_boolean(self.is_file_field, input_name='is-file-field'))
        if self.section is not None and 'section' not in already_processed:
            already_processed.add('section')
            outfile.write(' section=%s' % (self.gds_format_string(quote_attrib(self.section).encode(ExternalEncoding), input_name='section'), ))
        if self.is_forced_ontology is not None and 'is_forced_ontology' not in already_processed:
            already_processed.add('is_forced_ontology')
            outfile.write(' is-forced-ontology="%s"' % self.gds_format_boolean(self.is_forced_ontology, input_name='is-forced-ontology'))
        if self.header is not None and 'header' not in already_processed:
            already_processed.add('header')
            outfile.write(' header=%s' % (self.gds_format_string(quote_attrib(self.header).encode(ExternalEncoding), input_name='header'), ))
        if self.data_type is not None and 'data_type' not in already_processed:
            already_processed.add('data_type')
            outfile.write(' data-type=%s' % (self.gds_format_string(quote_attrib(self.data_type).encode(ExternalEncoding), input_name='data-type'), ))
        if self.is_multiple_value is not None and 'is_multiple_value' not in already_processed:
            already_processed.add('is_multiple_value')
            outfile.write(' is-multiple-value="%s"' % self.gds_format_boolean(self.is_multiple_value, input_name='is-multiple-value'))
        if self.is_hidden is not None and 'is_hidden' not in already_processed:
            already_processed.add('is_hidden')
            outfile.write(' is-hidden="%s"' % self.gds_format_boolean(self.is_hidden, input_name='is-hidden'))
        if self.is_required is not None and 'is_required' not in already_processed:
            already_processed.add('is_required')
            outfile.write(' is-required="%s"' % self.gds_format_boolean(self.is_required, input_name='is-required'))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='FieldType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.description is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdescription>%s</%sdescription>%s' % (namespace_, self.gds_format_string(quote_xml(self.description).encode(ExternalEncoding), input_name='description'), namespace_, eol_))
        if self.default_value is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdefault-value>%s</%sdefault-value>%s' % (namespace_, self.gds_format_string(quote_xml(self.default_value).encode(ExternalEncoding), input_name='default-value'), namespace_, eol_))
        if self.value_format is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%svalue-format>%s</%svalue-format>%s' % (namespace_, self.gds_format_string(quote_xml(self.value_format).encode(ExternalEncoding), input_name='value-format'), namespace_, eol_))
        if self.list_values is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%slist-values>%s</%slist-values>%s' % (namespace_, self.gds_format_string(quote_xml(self.list_values).encode(ExternalEncoding), input_name='list-values'), namespace_, eol_))
        if self.generated_value_template is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sgenerated-value-template>%s</%sgenerated-value-template>%s' % (namespace_, self.gds_format_string(quote_xml(self.generated_value_template).encode(ExternalEncoding), input_name='generated-value-template'), namespace_, eol_))
        if self.recommended_ontologies is not None:
            self.recommended_ontologies.export(outfile, level, namespace_='cfg:', name_='recommended-ontologies', pretty_print=pretty_print)
        if self.value_range is not None:
            self.value_range.export(outfile, level, namespace_, name_='value-range', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('is-file-field', node)
        if value is not None and 'is-file-field' not in already_processed:
            already_processed.add('is-file-field')
            if value in ('true', '1'):
                self.is_file_field = True
            elif value in ('false', '0'):
                self.is_file_field = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('section', node)
        if value is not None and 'section' not in already_processed:
            already_processed.add('section')
            self.section = value
        value = find_attr_value_('is-forced-ontology', node)
        if value is not None and 'is-forced-ontology' not in already_processed:
            already_processed.add('is-forced-ontology')
            if value in ('true', '1'):
                self.is_forced_ontology = True
            elif value in ('false', '0'):
                self.is_forced_ontology = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('header', node)
        if value is not None and 'header' not in already_processed:
            already_processed.add('header')
            self.header = value
        value = find_attr_value_('data-type', node)
        if value is not None and 'data-type' not in already_processed:
            already_processed.add('data-type')
            self.data_type = value
        value = find_attr_value_('is-multiple-value', node)
        if value is not None and 'is-multiple-value' not in already_processed:
            already_processed.add('is-multiple-value')
            if value in ('true', '1'):
                self.is_multiple_value = True
            elif value in ('false', '0'):
                self.is_multiple_value = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('is-hidden', node)
        if value is not None and 'is-hidden' not in already_processed:
            already_processed.add('is-hidden')
            if value in ('true', '1'):
                self.is_hidden = True
            elif value in ('false', '0'):
                self.is_hidden = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('is-required', node)
        if value is not None and 'is-required' not in already_processed:
            already_processed.add('is-required')
            if value in ('true', '1'):
                self.is_required = True
            elif value in ('false', '0'):
                self.is_required = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'description':
            description_ = child_.text
            description_ = self.gds_validate_string(description_, node, 'description')
            self.description = description_
        elif nodeName_ == 'default-value':
            default_value_ = child_.text
            default_value_ = self.gds_validate_string(default_value_, node, 'default_value')
            self.default_value = default_value_
        elif nodeName_ == 'value-format':
            value_format_ = child_.text
            value_format_ = self.gds_validate_string(value_format_, node, 'value_format')
            self.value_format = value_format_
        elif nodeName_ == 'list-values':
            list_values_ = child_.text
            list_values_ = self.gds_validate_string(list_values_, node, 'list_values')
            self.list_values = list_values_
        elif nodeName_ == 'generated-value-template':
            generated_value_template_ = child_.text
            generated_value_template_ = self.gds_validate_string(generated_value_template_, node, 'generated_value_template')
            self.generated_value_template = generated_value_template_
        elif nodeName_ == 'recommended-ontologies':
            obj_ = RecommendedOntologiesType.factory()
            obj_.build(child_)
            self.recommended_ontologies = obj_
            obj_.original_tagname_ = 'recommended-ontologies'
        elif nodeName_ == 'value-range':
            obj_ = ValueRangeType.factory()
            obj_.build(child_)
            self.value_range = obj_
            obj_.original_tagname_ = 'value-range'
# end class FieldType


class RecommendedOntologiesType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, ontology=None):
        self.original_tagname_ = None
        if ontology is None:
            self.ontology = []
        else:
            self.ontology = ontology
    def factory(*args_, **kwargs_):
        if RecommendedOntologiesType.subclass:
            return RecommendedOntologiesType.subclass(*args_, **kwargs_)
        else:
            return RecommendedOntologiesType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_ontology(self): return self.ontology
    def set_ontology(self, ontology): self.ontology = ontology
    def add_ontology(self, value): self.ontology.append(value)
    def insert_ontology_at(self, index, value): self.ontology.insert(index, value)
    def replace_ontology_at(self, index, value): self.ontology[index] = value
    def hasContent_(self):
        if (
            self.ontology
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='RecommendedOntologiesType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='RecommendedOntologiesType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='RecommendedOntologiesType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='RecommendedOntologiesType'):
        pass
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='RecommendedOntologiesType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for ontology_ in self.ontology:
            ontology_.export(outfile, level, namespace_, name_='ontology', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        pass
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'ontology':
            obj_ = OntologyType.factory()
            obj_.build(child_)
            self.ontology.append(obj_)
            obj_.original_tagname_ = 'ontology'
# end class RecommendedOntologiesType


class OntologyType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, abbreviation=None, version=None, id=None, name=None, branch=None):
        self.original_tagname_ = None
        self.abbreviation = _cast(None, abbreviation)
        self.version = _cast(None, version)
        self.id = _cast(None, id)
        self.name = _cast(None, name)
        if branch is None:
            self.branch = []
        else:
            self.branch = branch
    def factory(*args_, **kwargs_):
        if OntologyType.subclass:
            return OntologyType.subclass(*args_, **kwargs_)
        else:
            return OntologyType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_branch(self): return self.branch
    def set_branch(self, branch): self.branch = branch
    def add_branch(self, value): self.branch.append(value)
    def insert_branch_at(self, index, value): self.branch.insert(index, value)
    def replace_branch_at(self, index, value): self.branch[index] = value
    def get_abbreviation(self): return self.abbreviation
    def set_abbreviation(self, abbreviation): self.abbreviation = abbreviation
    def get_version(self): return self.version
    def set_version(self, version): self.version = version
    def get_id(self): return self.id
    def set_id(self, id): self.id = id
    def get_name(self): return self.name
    def set_name(self, name): self.name = name
    def hasContent_(self):
        if (
            self.branch
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='OntologyType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='OntologyType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='OntologyType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='OntologyType'):
        if self.abbreviation is not None and 'abbreviation' not in already_processed:
            already_processed.add('abbreviation')
            outfile.write(' abbreviation=%s' % (self.gds_format_string(quote_attrib(self.abbreviation).encode(ExternalEncoding), input_name='abbreviation'), ))
        if self.version is not None and 'version' not in already_processed:
            already_processed.add('version')
            outfile.write(' version=%s' % (self.gds_format_string(quote_attrib(self.version).encode(ExternalEncoding), input_name='version'), ))
        if self.id is not None and 'id' not in already_processed:
            already_processed.add('id')
            outfile.write(' id=%s' % (self.gds_format_string(quote_attrib(self.id).encode(ExternalEncoding), input_name='id'), ))
        if self.name is not None and 'name' not in already_processed:
            already_processed.add('name')
            outfile.write(' name=%s' % (self.gds_format_string(quote_attrib(self.name).encode(ExternalEncoding), input_name='name'), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='OntologyType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for branch_ in self.branch:
            branch_.export(outfile, level, namespace_, name_='branch', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('abbreviation', node)
        if value is not None and 'abbreviation' not in already_processed:
            already_processed.add('abbreviation')
            self.abbreviation = value
        value = find_attr_value_('version', node)
        if value is not None and 'version' not in already_processed:
            already_processed.add('version')
            self.version = value
        value = find_attr_value_('id', node)
        if value is not None and 'id' not in already_processed:
            already_processed.add('id')
            self.id = value
        value = find_attr_value_('name', node)
        if value is not None and 'name' not in already_processed:
            already_processed.add('name')
            self.name = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'branch':
            obj_ = BranchType.factory()
            obj_.build(child_)
            self.branch.append(obj_)
            obj_.original_tagname_ = 'branch'
# end class OntologyType


class BranchType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, id=None, name=None):
        self.original_tagname_ = None
        self.id = _cast(None, id)
        self.name = _cast(None, name)
    def factory(*args_, **kwargs_):
        if BranchType.subclass:
            return BranchType.subclass(*args_, **kwargs_)
        else:
            return BranchType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_id(self): return self.id
    def set_id(self, id): self.id = id
    def get_name(self): return self.name
    def set_name(self, name): self.name = name
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='BranchType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='BranchType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='BranchType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='BranchType'):
        if self.id is not None and 'id' not in already_processed:
            already_processed.add('id')
            outfile.write(' id=%s' % (self.gds_format_string(quote_attrib(self.id).encode(ExternalEncoding), input_name='id'), ))
        if self.name is not None and 'name' not in already_processed:
            already_processed.add('name')
            outfile.write(' name=%s' % (self.gds_format_string(quote_attrib(self.name).encode(ExternalEncoding), input_name='name'), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='BranchType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('id', node)
        if value is not None and 'id' not in already_processed:
            already_processed.add('id')
            self.id = value
        value = find_attr_value_('name', node)
        if value is not None and 'name' not in already_processed:
            already_processed.add('name')
            self.name = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class BranchType


class StructuredFieldType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, name=None):
        self.original_tagname_ = None
        self.name = _cast(None, name)
    def factory(*args_, **kwargs_):
        if StructuredFieldType.subclass:
            return StructuredFieldType.subclass(*args_, **kwargs_)
        else:
            return StructuredFieldType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_name(self): return self.name
    def set_name(self, name): self.name = name
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='StructuredFieldType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='StructuredFieldType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='StructuredFieldType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='StructuredFieldType'):
        if self.name is not None and 'name' not in already_processed:
            already_processed.add('name')
            outfile.write(' name=%s' % (self.gds_format_string(quote_attrib(self.name).encode(ExternalEncoding), input_name='name'), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='StructuredFieldType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('name', node)
        if value is not None and 'name' not in already_processed:
            already_processed.add('name')
            self.name = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class StructuredFieldType


class ProtocolFieldType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, data_type=None, protocol_type=None, is_required=None):
        self.original_tagname_ = None
        self.data_type = _cast(None, data_type)
        self.protocol_type = _cast(None, protocol_type)
        self.is_required = _cast(bool, is_required)
    def factory(*args_, **kwargs_):
        if ProtocolFieldType.subclass:
            return ProtocolFieldType.subclass(*args_, **kwargs_)
        else:
            return ProtocolFieldType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_data_type(self): return self.data_type
    def set_data_type(self, data_type): self.data_type = data_type
    def get_protocol_type(self): return self.protocol_type
    def set_protocol_type(self, protocol_type): self.protocol_type = protocol_type
    def get_is_required(self): return self.is_required
    def set_is_required(self, is_required): self.is_required = is_required
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='ProtocolFieldType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='ProtocolFieldType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='ProtocolFieldType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='ProtocolFieldType'):
        if self.data_type is not None and 'data_type' not in already_processed:
            already_processed.add('data_type')
            outfile.write(' data-type=%s' % (self.gds_format_string(quote_attrib(self.data_type).encode(ExternalEncoding), input_name='data-type'), ))
        if self.protocol_type is not None and 'protocol_type' not in already_processed:
            already_processed.add('protocol_type')
            outfile.write(' protocol-type=%s' % (self.gds_format_string(quote_attrib(self.protocol_type).encode(ExternalEncoding), input_name='protocol-type'), ))
        if self.is_required is not None and 'is_required' not in already_processed:
            already_processed.add('is_required')
            outfile.write(' is-required="%s"' % self.gds_format_boolean(self.is_required, input_name='is-required'))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='ProtocolFieldType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('data-type', node)
        if value is not None and 'data-type' not in already_processed:
            already_processed.add('data-type')
            self.data_type = value
        value = find_attr_value_('protocol-type', node)
        if value is not None and 'protocol-type' not in already_processed:
            already_processed.add('protocol-type')
            self.protocol_type = value
        value = find_attr_value_('is-required', node)
        if value is not None and 'is-required' not in already_processed:
            already_processed.add('is-required')
            if value in ('true', '1'):
                self.is_required = True
            elif value in ('false', '0'):
                self.is_required = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class ProtocolFieldType


class UnitFieldType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, is_multiple_value=None, data_type=None, is_required=None, is_forced_ontology=None, description=None, list_values=None, recommended_ontologies=None, default_value=None):
        self.original_tagname_ = None
        self.is_multiple_value = _cast(bool, is_multiple_value)
        self.data_type = _cast(None, data_type)
        self.is_required = _cast(bool, is_required)
        self.is_forced_ontology = _cast(bool, is_forced_ontology)
        self.description = description
        self.list_values = list_values
        self.recommended_ontologies = recommended_ontologies
        self.default_value = default_value
    def factory(*args_, **kwargs_):
        if UnitFieldType.subclass:
            return UnitFieldType.subclass(*args_, **kwargs_)
        else:
            return UnitFieldType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_description(self): return self.description
    def set_description(self, description): self.description = description
    def get_list_values(self): return self.list_values
    def set_list_values(self, list_values): self.list_values = list_values
    def get_recommended_ontologies(self): return self.recommended_ontologies
    def set_recommended_ontologies(self, recommended_ontologies): self.recommended_ontologies = recommended_ontologies
    def get_default_value(self): return self.default_value
    def set_default_value(self, default_value): self.default_value = default_value
    def get_is_multiple_value(self): return self.is_multiple_value
    def set_is_multiple_value(self, is_multiple_value): self.is_multiple_value = is_multiple_value
    def get_data_type(self): return self.data_type
    def set_data_type(self, data_type): self.data_type = data_type
    def get_is_required(self): return self.is_required
    def set_is_required(self, is_required): self.is_required = is_required
    def get_is_forced_ontology(self): return self.is_forced_ontology
    def set_is_forced_ontology(self, is_forced_ontology): self.is_forced_ontology = is_forced_ontology
    def hasContent_(self):
        if (
            self.description is not None or
            self.list_values is not None or
            self.recommended_ontologies is not None or
            self.default_value is not None
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='UnitFieldType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='UnitFieldType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='UnitFieldType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='UnitFieldType'):
        if self.is_multiple_value is not None and 'is_multiple_value' not in already_processed:
            already_processed.add('is_multiple_value')
            outfile.write(' is-multiple-value="%s"' % self.gds_format_boolean(self.is_multiple_value, input_name='is-multiple-value'))
        if self.data_type is not None and 'data_type' not in already_processed:
            already_processed.add('data_type')
            outfile.write(' data-type=%s' % (self.gds_format_string(quote_attrib(self.data_type).encode(ExternalEncoding), input_name='data-type'), ))
        if self.is_required is not None and 'is_required' not in already_processed:
            already_processed.add('is_required')
            outfile.write(' is-required="%s"' % self.gds_format_boolean(self.is_required, input_name='is-required'))
        if self.is_forced_ontology is not None and 'is_forced_ontology' not in already_processed:
            already_processed.add('is_forced_ontology')
            outfile.write(' is-forced-ontology="%s"' % self.gds_format_boolean(self.is_forced_ontology, input_name='is-forced-ontology'))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='UnitFieldType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.description is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdescription>%s</%sdescription>%s' % (namespace_, self.gds_format_string(quote_xml(self.description).encode(ExternalEncoding), input_name='description'), namespace_, eol_))
        if self.list_values is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%slist-values>%s</%slist-values>%s' % (namespace_, self.gds_format_string(quote_xml(self.list_values).encode(ExternalEncoding), input_name='list-values'), namespace_, eol_))
        if self.recommended_ontologies is not None:
            self.recommended_ontologies.export(outfile, level, namespace_='cfg:', name_='recommended-ontologies', pretty_print=pretty_print)
        if self.default_value is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdefault-value>%s</%sdefault-value>%s' % (namespace_, self.gds_format_string(quote_xml(self.default_value).encode(ExternalEncoding), input_name='default-value'), namespace_, eol_))
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('is-multiple-value', node)
        if value is not None and 'is-multiple-value' not in already_processed:
            already_processed.add('is-multiple-value')
            if value in ('true', '1'):
                self.is_multiple_value = True
            elif value in ('false', '0'):
                self.is_multiple_value = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('data-type', node)
        if value is not None and 'data-type' not in already_processed:
            already_processed.add('data-type')
            self.data_type = value
        value = find_attr_value_('is-required', node)
        if value is not None and 'is-required' not in already_processed:
            already_processed.add('is-required')
            if value in ('true', '1'):
                self.is_required = True
            elif value in ('false', '0'):
                self.is_required = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
        value = find_attr_value_('is-forced-ontology', node)
        if value is not None and 'is-forced-ontology' not in already_processed:
            already_processed.add('is-forced-ontology')
            if value in ('true', '1'):
                self.is_forced_ontology = True
            elif value in ('false', '0'):
                self.is_forced_ontology = False
            else:
                raise_parse_error(node, 'Bad boolean attribute')
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'description':
            description_ = child_.text
            description_ = self.gds_validate_string(description_, node, 'description')
            self.description = description_
        elif nodeName_ == 'list-values':
            list_values_ = child_.text
            list_values_ = self.gds_validate_string(list_values_, node, 'list_values')
            self.list_values = list_values_
        elif nodeName_ == 'recommended-ontologies':
            obj_ = RecommendedOntologiesType.factory()
            obj_.build(child_)
            self.recommended_ontologies = obj_
            obj_.original_tagname_ = 'recommended-ontologies'
        elif nodeName_ == 'default-value':
            default_value_ = child_.text
            default_value_ = self.gds_validate_string(default_value_, node, 'default_value')
            self.default_value = default_value_
# end class UnitFieldType


class ValueRangeType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, max=None, type_=None, min=None):
        self.original_tagname_ = None
        self.max = _cast(None, max)
        self.type_ = _cast(None, type_)
        self.min = _cast(None, min)
    def factory(*args_, **kwargs_):
        if ValueRangeType.subclass:
            return ValueRangeType.subclass(*args_, **kwargs_)
        else:
            return ValueRangeType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_max(self): return self.max
    def set_max(self, max): self.max = max
    def get_type(self): return self.type_
    def set_type(self, type_): self.type_ = type_
    def get_min(self): return self.min
    def set_min(self, min): self.min = min
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='ValueRangeType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='ValueRangeType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='ValueRangeType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='ValueRangeType'):
        if self.max is not None and 'max' not in already_processed:
            already_processed.add('max')
            outfile.write(' max=%s' % (self.gds_format_string(quote_attrib(self.max).encode(ExternalEncoding), input_name='max'), ))
        if self.type_ is not None and 'type_' not in already_processed:
            already_processed.add('type_')
            outfile.write(' type=%s' % (quote_attrib(self.type_), ))
        if self.min is not None and 'min' not in already_processed:
            already_processed.add('min')
            outfile.write(' min=%s' % (self.gds_format_string(quote_attrib(self.min).encode(ExternalEncoding), input_name='min'), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='ValueRangeType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('max', node)
        if value is not None and 'max' not in already_processed:
            already_processed.add('max')
            self.max = value
        value = find_attr_value_('type', node)
        if value is not None and 'type' not in already_processed:
            already_processed.add('type')
            self.type_ = value
        value = find_attr_value_('min', node)
        if value is not None and 'min' not in already_processed:
            already_processed.add('min')
            self.min = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class ValueRangeType


class IsaTabConfigurationType(GeneratedsSuper):
    """The conversion target. This is used by the converter to decide if a
    certain assay type can be converted into an omics-specific
    format or not. Current supported values are: magetab, prideml,
    ena. More targets can be added by extending the converter."""
    subclass = None
    superclass = None
    def __init__(self, table_name=None, isatab_conversion_target=None, isatab_assay_type=None, measurement=None, technology=None, field=None, protocol_field=None, structured_field=None, unit_field=None):
        self.original_tagname_ = None
        self.table_name = _cast(None, table_name)
        self.isatab_conversion_target = _cast(None, isatab_conversion_target)
        self.isatab_assay_type = _cast(None, isatab_assay_type)
        self.measurement = measurement
        self.technology = technology
        if field is None:
            self.field = []
        else:
            self.field = field
        if protocol_field is None:
            self.protocol_field = []
        else:
            self.protocol_field = protocol_field
        if structured_field is None:
            self.structured_field = []
        else:
            self.structured_field = structured_field
        if unit_field is None:
            self.unit_field = []
        else:
            self.unit_field = unit_field
    def factory(*args_, **kwargs_):
        if IsaTabConfigurationType.subclass:
            return IsaTabConfigurationType.subclass(*args_, **kwargs_)
        else:
            return IsaTabConfigurationType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_measurement(self): return self.measurement
    def set_measurement(self, measurement): self.measurement = measurement
    def get_technology(self): return self.technology
    def set_technology(self, technology): self.technology = technology
    def get_field(self): return self.field
    def set_field(self, field): self.field = field
    def add_field(self, value): self.field.append(value)
    def insert_field_at(self, index, value): self.field.insert(index, value)
    def replace_field_at(self, index, value): self.field[index] = value
    def get_protocol_field(self): return self.protocol_field
    def set_protocol_field(self, protocol_field): self.protocol_field = protocol_field
    def add_protocol_field(self, value): self.protocol_field.append(value)
    def insert_protocol_field_at(self, index, value): self.protocol_field.insert(index, value)
    def replace_protocol_field_at(self, index, value): self.protocol_field[index] = value
    def get_structured_field(self): return self.structured_field
    def set_structured_field(self, structured_field): self.structured_field = structured_field
    def add_structured_field(self, value): self.structured_field.append(value)
    def insert_structured_field_at(self, index, value): self.structured_field.insert(index, value)
    def replace_structured_field_at(self, index, value): self.structured_field[index] = value
    def get_unit_field(self): return self.unit_field
    def set_unit_field(self, unit_field): self.unit_field = unit_field
    def add_unit_field(self, value): self.unit_field.append(value)
    def insert_unit_field_at(self, index, value): self.unit_field.insert(index, value)
    def replace_unit_field_at(self, index, value): self.unit_field[index] = value
    def get_table_name(self): return self.table_name
    def set_table_name(self, table_name): self.table_name = table_name
    def get_isatab_conversion_target(self): return self.isatab_conversion_target
    def set_isatab_conversion_target(self, isatab_conversion_target): self.isatab_conversion_target = isatab_conversion_target
    def get_isatab_assay_type(self): return self.isatab_assay_type
    def set_isatab_assay_type(self, isatab_assay_type): self.isatab_assay_type = isatab_assay_type
    def hasContent_(self):
        if (
            self.measurement is not None or
            self.technology is not None or
            self.field or
            self.protocol_field or
            self.structured_field or
            self.unit_field
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='IsaTabConfigurationType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='IsaTabConfigurationType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='IsaTabConfigurationType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='IsaTabConfigurationType'):
        if self.table_name is not None and 'table_name' not in already_processed:
            already_processed.add('table_name')
            outfile.write(' table-name=%s' % (self.gds_format_string(quote_attrib(self.table_name).encode(ExternalEncoding), input_name='table-name'), ))
        if self.isatab_conversion_target is not None and 'isatab_conversion_target' not in already_processed:
            already_processed.add('isatab_conversion_target')
            outfile.write(' isatab-conversion-target=%s' % (self.gds_format_string(quote_attrib(self.isatab_conversion_target).encode(ExternalEncoding), input_name='isatab-conversion-target'), ))
        if self.isatab_assay_type is not None and 'isatab_assay_type' not in already_processed:
            already_processed.add('isatab_assay_type')
            outfile.write(' isatab-assay-type=%s' % (quote_attrib(self.isatab_assay_type), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='IsaTabConfigurationType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.measurement is not None:
            self.measurement.export(outfile, level, namespace_, name_='measurement', pretty_print=pretty_print)
        if self.technology is not None:
            self.technology.export(outfile, level, namespace_, name_='technology', pretty_print=pretty_print)
        for field_ in self.field:
            field_.export(outfile, level, namespace_='cfg:', name_='field', pretty_print=pretty_print)
        for protocol_field_ in self.protocol_field:
            protocol_field_.export(outfile, level, namespace_='cfg:', name_='protocol-field', pretty_print=pretty_print)
        for structured_field_ in self.structured_field:
            structured_field_.export(outfile, level, namespace_='cfg:', name_='structured-field', pretty_print=pretty_print)
        for unit_field_ in self.unit_field:
            unit_field_.export(outfile, level, namespace_='cfg:', name_='unit-field', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        pos = 0
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_, pos)
            if not ((nodeName_ == 'measurement') or (nodeName_ == 'technology')):
                pos += 1
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('table-name', node)
        if value is not None and 'table-name' not in already_processed:
            already_processed.add('table-name')
            self.table_name = value
        value = find_attr_value_('isatab-conversion-target', node)
        if value is not None and 'isatab-conversion-target' not in already_processed:
            already_processed.add('isatab-conversion-target')
            self.isatab_conversion_target = value
        value = find_attr_value_('isatab-assay-type', node)
        if value is not None and 'isatab-assay-type' not in already_processed:
            already_processed.add('isatab-assay-type')
            self.isatab_assay_type = value
    def buildChildren(self, child_, node, nodeName_, pos, fromsubclass_=False):
        if nodeName_ == 'measurement':
            obj_ = OntologyEntryType.factory()
            obj_.build(child_)
            self.measurement = obj_
            obj_.original_tagname_ = 'measurement'
        elif nodeName_ == 'technology':
            obj_ = OntologyEntryType.factory()
            obj_.build(child_)
            self.technology = obj_
            obj_.original_tagname_ = 'technology'
        elif nodeName_ == 'field':
            obj_ = FieldType.factory()
            obj_.build(child_)
            self.field.append(obj_)
            obj_.original_tagname_ = 'field'
            obj_.pos = pos
        elif nodeName_ == 'protocol-field':
            obj_ = ProtocolFieldType.factory()
            obj_.build(child_)
            self.protocol_field.append(obj_)
            obj_.original_tagname_ = 'protocol-field'
            obj_.pos = pos
        elif nodeName_ == 'structured-field':
            obj_ = StructuredFieldType.factory()
            obj_.build(child_)
            self.structured_field.append(obj_)
            obj_.original_tagname_ = 'structured-field'
            obj_.pos = pos
        elif nodeName_ == 'unit-field':
            obj_ = UnitFieldType.factory()
            obj_.build(child_)
            self.unit_field.append(obj_)
            obj_.original_tagname_ = 'unit-field'
            obj_.pos = pos
# end class IsaTabConfigurationType


class IsaTabConfigFileType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, isatab_configuration=None):
        self.original_tagname_ = None
        if isatab_configuration is None:
            self.isatab_configuration = []
        else:
            self.isatab_configuration = isatab_configuration
    def factory(*args_, **kwargs_):
        if IsaTabConfigFileType.subclass:
            return IsaTabConfigFileType.subclass(*args_, **kwargs_)
        else:
            return IsaTabConfigFileType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_isatab_configuration(self): return self.isatab_configuration
    def set_isatab_configuration(self, isatab_configuration): self.isatab_configuration = isatab_configuration
    def add_isatab_configuration(self, value): self.isatab_configuration.append(value)
    def insert_isatab_configuration_at(self, index, value): self.isatab_configuration.insert(index, value)
    def replace_isatab_configuration_at(self, index, value): self.isatab_configuration[index] = value
    def hasContent_(self):
        if (
            self.isatab_configuration
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='IsaTabConfigFileType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='IsaTabConfigFileType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='IsaTabConfigFileType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='IsaTabConfigFileType'):
        pass
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='IsaTabConfigFileType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for isatab_configuration_ in self.isatab_configuration:
            isatab_configuration_.export(outfile, level, namespace_='cfg:', name_='isatab-configuration', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        pass
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'isatab-configuration':
            obj_ = IsaTabConfigurationType.factory()
            obj_.build(child_)
            self.isatab_configuration.append(obj_)
            obj_.original_tagname_ = 'isatab-configuration'
# end class IsaTabConfigFileType


class OntologyEntryType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, term_accession=None, term_label=None, source_version=None, source_title=None, source_abbreviation=None, source_uri=None):
        self.original_tagname_ = None
        self.term_accession = _cast(None, term_accession)
        self.term_label = _cast(None, term_label)
        self.source_version = _cast(None, source_version)
        self.source_title = _cast(None, source_title)
        self.source_abbreviation = _cast(None, source_abbreviation)
        self.source_uri = _cast(None, source_uri)
    def factory(*args_, **kwargs_):
        if OntologyEntryType.subclass:
            return OntologyEntryType.subclass(*args_, **kwargs_)
        else:
            return OntologyEntryType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_term_accession(self): return self.term_accession
    def set_term_accession(self, term_accession): self.term_accession = term_accession
    def get_term_label(self): return self.term_label
    def set_term_label(self, term_label): self.term_label = term_label
    def get_source_version(self): return self.source_version
    def set_source_version(self, source_version): self.source_version = source_version
    def get_source_title(self): return self.source_title
    def set_source_title(self, source_title): self.source_title = source_title
    def get_source_abbreviation(self): return self.source_abbreviation
    def set_source_abbreviation(self, source_abbreviation): self.source_abbreviation = source_abbreviation
    def get_source_uri(self): return self.source_uri
    def set_source_uri(self, source_uri): self.source_uri = source_uri
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='cfg:', name_='OntologyEntryType', namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='OntologyEntryType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='cfg:', name_='OntologyEntryType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='cfg:', name_='OntologyEntryType'):
        if self.term_accession is not None and 'term_accession' not in already_processed:
            already_processed.add('term_accession')
            outfile.write(' term-accession=%s' % (self.gds_format_string(quote_attrib(self.term_accession).encode(ExternalEncoding), input_name='term-accession'), ))
        if self.term_label is not None and 'term_label' not in already_processed:
            already_processed.add('term_label')
            outfile.write(' term-label=%s' % (self.gds_format_string(quote_attrib(self.term_label).encode(ExternalEncoding), input_name='term-label'), ))
        if self.source_version is not None and 'source_version' not in already_processed:
            already_processed.add('source_version')
            outfile.write(' source-version=%s' % (self.gds_format_string(quote_attrib(self.source_version).encode(ExternalEncoding), input_name='source-version'), ))
        if self.source_title is not None and 'source_title' not in already_processed:
            already_processed.add('source_title')
            outfile.write(' source-title=%s' % (self.gds_format_string(quote_attrib(self.source_title).encode(ExternalEncoding), input_name='source-title'), ))
        if self.source_abbreviation is not None and 'source_abbreviation' not in already_processed:
            already_processed.add('source_abbreviation')
            outfile.write(' source-abbreviation=%s' % (self.gds_format_string(quote_attrib(self.source_abbreviation).encode(ExternalEncoding), input_name='source-abbreviation'), ))
        if self.source_uri is not None and 'source_uri' not in already_processed:
            already_processed.add('source_uri')
            outfile.write(' source-uri=%s' % (self.gds_format_string(quote_attrib(self.source_uri).encode(ExternalEncoding), input_name='source-uri'), ))
    def exportChildren(self, outfile, level, namespace_='cfg:', name_='OntologyEntryType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('term-accession', node)
        if value is not None and 'term-accession' not in already_processed:
            already_processed.add('term-accession')
            self.term_accession = value
        value = find_attr_value_('term-label', node)
        if value is not None and 'term-label' not in already_processed:
            already_processed.add('term-label')
            self.term_label = value
        value = find_attr_value_('source-version', node)
        if value is not None and 'source-version' not in already_processed:
            already_processed.add('source-version')
            self.source_version = value
        value = find_attr_value_('source-title', node)
        if value is not None and 'source-title' not in already_processed:
            already_processed.add('source-title')
            self.source_title = value
        value = find_attr_value_('source-abbreviation', node)
        if value is not None and 'source-abbreviation' not in already_processed:
            already_processed.add('source-abbreviation')
            self.source_abbreviation = value
        value = find_attr_value_('source-uri', node)
        if value is not None and 'source-uri' not in already_processed:
            already_processed.add('source-uri')
            self.source_uri = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class OntologyEntryType


GDSClassesMapping = {
    'isatab-config-file': IsaTabConfigFileType,
    'recommended-ontologies': RecommendedOntologiesType,
    'structured-field': StructuredFieldType,
    'unit-field': UnitFieldType,
    'isatab-configuration': IsaTabConfigurationType,
    'field': FieldType,
    'branch': BranchType,
    'measurement': OntologyEntryType,
    'value-range': ValueRangeType,
    'protocol-field': ProtocolFieldType,
    'ontology': OntologyType,
    'technology': OntologyEntryType,
}


USAGE_TEXT = """
Usage: python <Parser>.py [ -s ] <in_xml_file>
"""


def usage():
    print(USAGE_TEXT)
    sys.exit(1)


def get_root_tag(node):
    tag = Tag_pattern_.match(node.tag).groups()[-1]
    rootClass = GDSClassesMapping.get(tag)
    if rootClass is None:
        rootClass = globals().get(tag)
    return tag, rootClass


def parse(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'FieldType'
        rootClass = FieldType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"',
            pretty_print=True)
    return rootObj


def parseEtree(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'FieldType'
        rootClass = FieldType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    mapping = {}
    rootElement = rootObj.to_etree(None, name_=rootTag, mapping_=mapping)
    reverse_mapping = rootObj.gds_reverse_node_mapping(mapping)
    if not silence:
        content = etree_.tostring(
            rootElement, pretty_print=True,
            xml_declaration=True, encoding="utf-8")
        sys.stdout.write(content)
        sys.stdout.write('\n')
    return rootObj, rootElement, mapping, reverse_mapping


def parseString(inString, silence=False):
    from io import StringIO
    parser = None
    doc = parsexml_(StringIO(inString), parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'FieldType'
        rootClass = FieldType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='xmlns:cfg="http://www.ebi.ac.uk/bii/isatab_configuration#"')
    return rootObj


def parseLiteral(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'FieldType'
        rootClass = FieldType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('#from isatab_config import *\n\n')
        sys.stdout.write('import isatab_config as model_\n\n')
        sys.stdout.write('rootObj = model_.rootClass(\n')
        rootObj.exportLiteral(sys.stdout, 0, name_=rootTag)
        sys.stdout.write(')\n')
    return rootObj


def main():
    args = sys.argv[1:]
    if len(args) == 1:
        parse(args[0])
    else:
        usage()


if __name__ == '__main__':
    #import pdb; pdb.set_trace()
    main()


__all__ = [
    "BranchType",
    "FieldType",
    "IsaTabConfigFileType",
    "IsaTabConfigurationType",
    "OntologyEntryType",
    "OntologyType",
    "ProtocolFieldType",
    "RecommendedOntologiesType",
    "StructuredFieldType",
    "UnitFieldType",
    "ValueRangeType"
]



# -*- coding: utf-8 -*
"""Convert ISA-Tab to SampleTab"""
import logging

from isatools import isatab, sampletab


log = logging.getLogger('isatools')


def convert(source_inv_fp, target_fp):
    """ Converter for ISA-Tab to SampleTab.
    :param source_inv_fp: File descriptor of input investigation file
    :param target_fp: File descriptor to write output SampleTab to (must be
                      writeable)
    """
    log.info("loading isatab %s", source_inv_fp.name)
    ISA = isatab.load(source_inv_fp)
    log.info("dumping sampletab %s", target_fp.name)
    sampletab.dump(ISA, target_fp)



# -*- coding: utf-8 -*-
"""Functions for querying PubMed.

This module connects to the PubMed API via Entrez.
If you have problems with it, check that it's working at
https://www.ncbi.nlm.nih.gov/pubmed/
"""
from __future__ import absolute_import
import logging

from Bio import Entrez, Medline

from isatools.model import Comment, Publication


log = logging.getLogger('isatools')


def get_pubmed_article(pubmed_id):
    # http://biopython.org/DIST/docs/tutorial/Tutorial.html#htoc126
    response = {}
    Entrez.email = "isatools@googlegroups.com"
    handle = Entrez.efetch(db="pubmed", id=pubmed_id.strip(),
                           rettype="medline", retmode="text")
    records = Medline.parse(handle)
    for record in records:
        response["pubmedid"] = pubmed_id
        response["title"] = record.get("TI", "")
        response["authors"] = record.get("AU", "")
        response["journal"] = record.get("TA", "")
        response["year"] = record.get("EDAT", "").split("/")[0]
        lidstring = record.get("LID", "")
        if "[doi]" in lidstring:
            response["doi"] = record.get("LID", "").split(" ")[0]
        else:
            response["doi"] = ""
        if not response["doi"]:
            aids = record.get("AID", "")
            for aid in aids:
                log.debug("AID:" + aid)
                if "[doi]" in aid:
                    response["doi"] = aid.split(" ")[0]
                    break
                else:
                    response["doi"] = ""

        break
    return response


def set_pubmed_article(publication):
    """
        Given a Publication object with pubmed_id set to some value, set the
        rest of the values from information
        collected via Entrez webservice from PubMed
    """
    if isinstance(publication, Publication):
        response = get_pubmed_article(publication.pubmed_id)
        publication.doi = response["doi"]
        publication.author_list = ", ".join(response["authors"])
        publication.title = response["title"]
        publication.comments = [Comment(name="Journal",
                                        value=response["journal"])]
    else:
        raise TypeError("Can only set PubMed details on a Publication object")



"""Convert MAGE-TAB to ISA-Tab"""
import logging
import os

from isatools import isatab
from isatools.magetab import MageTabParser


log = logging.getLogger('isatools')


def convert(idf_file_path, output_path):
    """ Converter for MAGE-TAB to ISA-Tab
    :param idf_file_path: File descriptor of input IDF file
    :param output_path: Path to directory to write output ISA-Tab files to
    """
    parser = MageTabParser()
    parser.parse_idf(idf_file_path)
    sdrf_files = [x.value for x in parser.ISA.studies[-1].comments if
                  'SDRF File' in x.name]
    if len(sdrf_files) == 1:
        sdrf_files = sdrf_files[0].split(';')
        for sdrf_file in sdrf_files:
            table_files = parser.parse_sdrf_to_isa_table_files(
                os.path.join(os.path.dirname(idf_file_path), sdrf_file))
            for in_fp in table_files:
                log.info("Writing {0} to {1}".format(in_fp.name, output_path))
                with open(os.path.join(output_path, in_fp.name), 'w',
                          encoding='utf-8') as out_fp:
                    out_fp.write(in_fp.read())
    log.info("Writing {0} to {1}".format("i_investigation.txt", output_path))
    isatab.dump(parser.ISA, output_path=output_path, skip_dump_tables=True)
